; $Id: fis_temp.pro,v 1.2 2006/09/28 06:18:08 makiuti Exp $
;+
; NAME:
;   FISSW_TEMP
;
; PURPOSE:
;       This function receives FIS thermometer output in ADU,
;       then returns temperature values in absolute temperature.
;
;       =-----  For SW Detector  -----=;
;
;/* for Cernox S/N X25345
;
;Convert from ADU(=TSD output) to Resistance
;       Output Codes and Analog input conversion
;       (ADS7809 data sheet, Burr-Brown Products, 2003)
;       STRAIGT BINARY   +-5V 16bit
;       LSB                     153uV  (5 / 0x8000 = 5 / 32768 = 1.52587891e-4)
;       +Full Scale (FS - 1LSB) 0xFFFF   4.999847V  (65535)
;       Midscale                0x8000   0V         (32768 = k0VADU)
;       One LSB Below Midscale  0x7FFF   -153uV     (32767)
;       -FullScale              0x0000   -5V        (0)
;
;   16-bit:     0x0000    0x7FFF   0x8000   0xFFFF
;   TSD output:      0     32767   -32767       -1
;
; MODIFICATION HISTORY:
;     Revised on Mar.30,2004 By T.Hirao
;     Transplanted from C to IDL on April 8, 2006 By M.Shirahata
;     06/09/28 Last modified by Makiuti
;
;-
FUNCTION FISSW_TEMP, ADU, OFFSET=offset

IF n_elements(offset) EQ 0 THEN offset = 198 ; (based on FISDF-001.fits)

;ADU Offset adjustment
adu = adu + offset

;ADU --> Volt
volt = FltArr(n_elements(adu))
;
idx = where((adu GE 0) and (adu LE 32767), cnt)
IF (cnt GT 0) THEN volt[idx] = 1.52587891e-4 * adu[idx] - 5.0
idx = where((adu LE -1) and (adu GE -32767), cnt)
IF (cnt GT 0) THEN volt[idx] = 1.52587891e-4 * adu[idx] + 5.0

;Volt --> Resistance
;  Current = 1uA,  AMP = *1000
resistance = abs(double(volt))/1000.0/1e-6

;Define the correction table for thermometer(Cernox S/N X25345)
sw_ctable = dblarr(8,3)
sw_ctable = [[178.058286d, -127.893413d, 22.913649d, -3.706396d, 0.858673d, -0.193427d, 0.044182d, -0.014904d], $
              [43.071237d, -38.199009d, 7.887000d, -0.895183d, 0.094139d, 0.005097d, 0.0d, 0.0d], $
              [5.929053d, -6.652762d, 2.508323d, -0.697874d, 0.140454d, -0.016493d, -0.001222d, 0.001031d]]
zu = [1.95266959383236d, 2.35249762656815d, 2.9570103799292d]
zl = [1.51756643374661d, 1.88711844525893d, 2.29277252151184d]

;Initialization
x = 0.0d
temperature = make_array(n_elements(resistance), /double, value=0.0)

;Calculation
i = make_array(n_elements(resistance), /integer, value=-1) ; -1 means calibration out of range
idx = where((resistance GT 209.1) and (resistance LT 801.3), cnt)
IF (cnt GT 0) THEN i[idx] = 2
idx = where((resistance GT 82.82) and (resistance LT 209.1), cnt)
IF (cnt GT 0) THEN i[idx] = 1
idx = where((resistance GT 33.46) and (resistance LT 82.82), cnt)
IF (cnt GT 0) THEN i[idx] = 0

resistance = alog10(resistance)

FOR k = 0L, n_elements(resistance)-1 DO BEGIN
  IF (i[k] GE 0) THEN BEGIN
    x = ((resistance[k] - zl[i[k]]) - (zu[i[k]] - resistance[k])) / (zu[i[k]] - zl[i[k]])
    if (x GT 1) or (x LT -1) then begin
      temperature[k] = 0.0
    endif else begin
      for j=0L, 7 do begin
        temperature[k] = temperature[k] + sw_ctable[j,i[k]] * cos( double(j)*acos(x) )
      endfor
    endelse
  ENDIF ELSE BEGIN
    temperature[k] = 0.0
  ENDELSE
ENDFOR

  return, temperature
END

;+
; NAME:
;   FISLW_TEMP
;
; PURPOSE:
;       This function receives FIS thermometer output in ADU,
;       then returns temperature values in absolute temperature.
;
;       =-----  For LW Detector  -----=;
;
;/* for Cernox S/N X12297
;
;Convert from ADU(=TSD output) to Resistance
;       Output Codes and Analog input conversion
;       (ADS7809 data sheet, Burr-Brown Products, 2003)
;       STRAIGT BINARY   +-5V 16bit
;       LSB                     153uV  (5 / 0x8000 = 5 / 32768 = 1.52587891e-4)
;       +Full Scale (FS - 1LSB) 0xFFFF   4.999847V  (65535)
;       Midscale                0x8000   0V         (32768 = k0VADU)
;       One LSB Below Midscale  0x7FFF   -153uV     (32767)
;       -FullScale              0x0000   -5V        (0)
;
;   16-bit:     0x0000    0x7FFF   0x8000   0xFFFF
;   TSD output:      0     32767   -32767       -1
;
; MODIFICATION HISTORY:
;     Revised on Mar.30,2004 By T.Hirao
;     Transplanted from C to IDL on April 8, 2006 By M.Shirahata
;     06/09/28 Last modified by Makiuti
;
;-
FUNCTION FISLW_TEMP, ADU, OFFSET=offset

IF n_elements(offset) EQ 0 THEN offset = 230 ; (derived from data at 2006/09/23 ~16:00)

;ADU Offset adjustment
adu = adu + offset

;ADU --> Volt
volt = FltArr(n_elements(adu))
;
idx = where((adu GE 0) and (adu LE 32767), cnt)
IF (cnt GT 0) THEN volt[idx] = 1.52587891e-4 * adu[idx] - 5.0
idx = where((adu LE -1) and (adu GE -32767), cnt)
IF (cnt GT 0) THEN volt[idx] = 1.52587891e-4 * adu[idx] + 5.0

;Volt --> Resistance
;  Current = 1uA,  AMP = *1000
resistance = abs(double(volt))/1000.0/1e-6

;Define the correction table for thermometer(Cernox S/N X12297)
lw_ctable = dblarr(8,3)
lw_ctable = [[64.273689d, -47.131708d, 7.881123d, -0.742952d, 0.019693d, 0.004753d, 0.0d, 0.0d], $
              [15.143251d, -12.658632d, 2.873647d, -0.350415d, 0.004080d, 0.0d, 0.0d, 0.0d], $
              [3.588521d, -3.172419d, 0.978489d, -0.235176d, 0.045005d, -0.006893d, 0.0d, 0.0d]]
zu = [2.32674686866d, 2.71233723599d, 3.21155554968d]
zl = [1.88921101015d, 2.26609636323d, 2.59139357292d]

;Initialization
x = 0.0d
temperature = make_array(n_elements(resistance), /double, value=0.0)

;Calculation
i = make_array(n_elements(resistance), /integer, value=-1) ; -1 means calibration out of range
idx = where((resistance GT 452.4), cnt)
IF (cnt GT 0) THEN i[idx] = 2
idx = where((resistance GT 197.5) and (resistance LT 452.4), cnt)
IF (cnt GT 0) THEN i[idx] = 1
idx = where((resistance GT 87.99) and (resistance LT 197.5), cnt)
IF (cnt GT 0) THEN i[idx] = 0

resistance = alog10(resistance)

FOR k = 0L, n_elements(resistance)-1 DO BEGIN
  IF (i[k] GE 0) THEN BEGIN
    x = ((resistance[k] - zl[i[k]]) - (zu[i[k]] - resistance[k])) / (zu[i[k]] - zl[i[k]])
    if (x GT 1) or (x LT -1) then begin
      temperature[k] = 0.0
    endif else begin
      for j=0L, 7 do begin
        temperature[k] = temperature[k] + lw_ctable[j,i[k]] * cos( double(j)*acos(x) )
      endfor
    endelse
  ENDIF ELSE BEGIN
    temperature[k] = 0.0
  ENDELSE
ENDFOR

  return, temperature
END

;+
; NAME:
;   FISBODY_TEMP
;
; PURPOSE:
;       This function receives FIS thermometer output in ADU,
;       then returns temperature values in absolute temperature.
;
;       =-----  For FIS Body  -----=;
;
;/* for Cernox S/N X12998
;
;Convert from ADU(=TSD output) to Resistance
;       Output Codes and Analog input conversion
;       (ADS7809 data sheet, Burr-Brown Products, 2003)
;       STRAIGT BINARY   +-5V 16bit
;       LSB                     153uV  (5 / 0x8000 = 5 / 32768 = 1.52587891e-4)
;       +Full Scale (FS - 1LSB) 0xFFFF   4.999847V  (65535)
;       Midscale                0x8000   0V         (32768 = k0VADU)
;       One LSB Below Midscale  0x7FFF   -153uV     (32767)
;       -FullScale              0x0000   -5V        (0)
;
;   16-bit:     0x0000    0x7FFF   0x8000   0xFFFF
;   TSD output:      0     32767   -32767       -1
;
; MODIFICATION HISTORY:
;     Revised on Mar.30,2004 By T.Hirao
;     Transplanted from C to IDL on April 8, 2006 By M.Shirahata
;     06/09/28 Last modified by Makiuti
;
;-
FUNCTION FISBODY_TEMP, ADU, OFFSET=offset

IF n_elements(offset) EQ 0 THEN offset = 218; (derived from data at 2006/09/23 ~16:00)

;ADU Offset adjustment
adu = adu + offset

;ADU --> Volt
volt = FltArr(n_elements(adu))
;
idx = where((adu GE 0) and (adu LE 32767), cnt)
IF (cnt GT 0) THEN volt[idx] = 1.52587891e-4 * adu[idx] - 5.0
idx = where((adu LE -1) and (adu GE -32767), cnt)
IF (cnt GT 0) THEN volt[idx] = 1.52587891e-4 * adu[idx] + 5.0

;Volt --> Resistance
;  Current = 1uA,  AMP = *1000
resistance = abs(double(volt))/1000.0/1e-6

;Define the correction table for thermometer(Cernox S/N X12998)
fis_ctable = dblarr(8,3)
fis_ctable = [[64.300220d, -47.161290d, 7.887435d, -0.741674d, 0.020082d, 0.006352d, 0.0d, 0.0d], $
              [15.158464d, -12.653010d, 2.866127d, -0.347403d, 0.005286d, 0.002308d, 0.0d, 0.0d], $
              [3.604269d, -3.197230d, 0.990886d, -0.239398d, 0.045766d, -0.006546d, 0.0d, 0.0d]]
zu = [2.33296020224d, 2.72068655975d, 3.22590580548d]
zl = [1.89232811923d, 2.2718831465d, 2.597819683d]

;Initialization
x = 0.0d
temperature = make_array(n_elements(resistance), /double, value=0.0)

;Calculation
i = make_array(n_elements(resistance), /integer, value=-1) ; -1 means calibration out of range
idx = where((resistance GT 460.2), cnt)
IF (cnt GT 0) THEN i[idx] = 2
idx = where((resistance GT 200.3) and (resistance LT 460.2), cnt)
IF (cnt GT 0) THEN i[idx] = 1
idx = where((resistance GT 88.73) and (resistance LT 200.3), cnt)
IF (cnt GT 0) THEN i[idx] = 0

resistance = alog10(resistance)

FOR k = 0L, n_elements(resistance)-1 DO BEGIN
  IF (i[k] GE 0) THEN BEGIN
    x = ((resistance[k] - zl[i[k]]) - (zu[i[k]] - resistance[k])) / (zu[i[k]] - zl[i[k]])
    if (x GT 1) or (x LT -1) then begin
      temperature[k] = 0.0
    endif else begin
      for j=0L, 7 do begin
        temperature[k] = temperature[k] + fis_ctable[j,i[k]] * cos( double(j)*acos(x) )
      endfor
    endelse
  ENDIF ELSE BEGIN
    temperature[k] = 0.0
  ENDELSE
ENDFOR

  return, temperature
END

;+
; NAME:
;   FISFTS_TEMP
;
; PURPOSE:
;       This function receives FIS thermometer output in ADU,
;       then returns temperature values in absolute temperature.
;
;       =-----  For FTS (Matsuo Cal lamp)  -----=;
;
;/* for cernox  MT#4
;
;Convert from ADU(=TSD output) to Resistance
;       Output Codes and Analog input conversion
;       (ADS7809 data sheet, Burr-Brown Products, 2003)
;       STRAIGT BINARY   +-5V 16bit
;       LSB                     153uV  (5 / 0x8000 = 5 / 32768 = 1.52587891e-4)
;       +Full Scale (FS - 1LSB) 0xFFFF   4.999847V  (65535)
;       Midscale                0x8000   0V         (32768 = k0VADU)
;       One LSB Below Midscale  0x7FFF   -153uV     (32767)
;       -FullScale              0x0000   -5V        (0)
;
;   16-bit:     0x0000    0x7FFF   0x8000   0xFFFF
;   TSD output:      0     32767   -32767       -1
;
; MODIFICATION HISTORY:
;     Calibrated by H.Takahashi.
;     Fit was done by T.Hirao (2004/4/27)
;     Transplanted from C to IDL on Sep 25, 2006 By M.Shirahata
;     06/09/28 Last modified by Makiuti
;
;-
FUNCTION FISFTS_TEMP, ADU, OFFSET=offset

IF n_elements(offset) EQ 0 THEN offset = 405 ;

;ADU Offset adjustment
adu = adu + offset

;ADU --> Volt
volt = FltArr(n_elements(adu))
;
idx = where((adu GE 0) and (adu LE 32767), cnt)
IF (cnt GT 0) THEN volt[idx] = 1.52587891e-4 * adu[idx] - 5.0
idx = where((adu LE -1) and (adu GE -32767), cnt)
IF (cnt GT 0) THEN volt[idx] = 1.52587891e-4 * adu[idx] + 5.0

;Volt --> Resistance
;  Current = 1uA,  AMP = *1000
resistance = abs(double(volt))/1000.0/1e-6

;Define the correction table for thermometer()
fts_ctable1 = dblarr(6)
fts_ctable1 = [3.6369d, 3.8695d, 4.2201d, 4.5662d, 4.3295d, 4.5319d]
fts_ctable2 = dblarr(6)
fts_ctable2 = [-0.90346d, -0.97545d, -1.0881d, -1.2074d, -1.107d, -1.2033d]

;Initialization
temperature = make_array(n_elements(resistance), /double, value=0.0)
logT = make_array(n_elements(resistance), /double, value=0.0)

;Calculation
i = make_array(n_elements(resistance), /integer, value=-1) ; -1 means calibration out of range
idx = where((resistance GT 2166.9), cnt) ; T < 4.20K  Uncalibrated region
IF (cnt GT 0) THEN i[idx] = -1
idx = where((resistance GE 1684.7) and (resistance LT 2166.9), cnt) ; 4.20K < T < 5.28K
IF (cnt GT 0) THEN i[idx] = 0
idx = where((resistance GE 1316.0) and (resistance LT 1684.7), cnt) ; 5.28K < T < 6.72K
IF (cnt GT 0) THEN i[idx] = 1
idx = where((resistance GE 835.7) and (resistance LT 1316.0), cnt) ; 6.72K < T < 11.02K
IF (cnt GT 0) THEN i[idx] = 2
idx = where((resistance GE 259.8) and (resistance LT 835.7), cnt) ; 11.02K < T < 45.40K
IF (cnt GT 0) THEN i[idx] = 3
idx = where((resistance GE 153.0) and (resistance LT 259.8), cnt) ; 45.40K < T < 79.98K
IF (cnt GT 0) THEN i[idx] = 4
idx = where((resistance GE 51.0) and (resistance LT 153.0), cnt) ; 79.98K < T < 300K
IF (cnt GT 0) THEN i[idx] = 5
idx = where((resistance LT 51.0), cnt) ; 300K < T
IF (cnt GT 0) THEN i[idx] = -1

resistance = alog10(resistance)

FOR k = 0L, n_elements(resistance)-1 DO BEGIN
  IF (i[k] GE 0) THEN BEGIN
        logT[k] = fts_ctable1[i[k]] + fts_ctable2[i[k]] * resistance[k]
        temperature[k] = 10.0^(logT[k])
  ENDIF ELSE BEGIN
    temperature[k] = 0.0
  ENDELSE
ENDFOR

  return, temperature
END


;=====================================================================
;  MAIN FUNCTION
;=====================================================================
;+
; NAME:
;   FIS_TEMP
;
; PURPOSE:
;       This function receives FIS thermometer output in ADU,
;       then returns temperature values in absolute temperature.
;
; MODIFICATION HISTORY:
;   Written by Makiuti,S.
;   06/09/28 Last modified
;-
FUNCTION FIS_TEMP, THERMO, ADU, OFFSET=offset

  IF (n_params() LT 2) THEN thermo = 'DUMMY'

;print, Size(thermo, /Type), Arg_Present(thermo), thermo

  IF (thermo EQ 'SW') THEN BEGIN
    ret = FISSW_TEMP(adu, offset=offset)
  ENDIF ELSE IF (thermo EQ 'LW') THEN BEGIN
    ret = FISLW_TEMP(adu, offset=offset)
  ENDIF ELSE IF (thermo EQ 'BODY') THEN BEGIN
    ret = FISBODY_TEMP(adu, offset=offset)
  ENDIF ELSE IF (thermo EQ 'FTS') THEN BEGIN
    ret = FISFTS_TEMP(adu, offset=offset)
  ENDIF ELSE BEGIN
    print, "'SW', 'LW' 'BODY', and 'FTS' are allowed as an argument 'THERMO'"
    print, "array = FIS_TEMP(THERMO, ADU, OFFSET=offset)"
    print, "ex.) IDL> ret = fis_temp('SW', 29127)"
    return, -1
  ENDELSE

  return ,ret

END
