; $Id: gb_periodic_noise.pro,v 1.2 2006/10/10 12:30:16 yamamura Exp $
;
; Copyright (c) 2002, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;	GB_periodic_noise.pro
; PURPOSE:
;	detect and correct for periodic noise
; CATEGORY:
;	Function. Green Box procedure in ASTRO-F pipeline.
; CALLING SEQUENCE:
;	Result = GB_periodic_noise( tsd, [Calib_Const_File = ccf, $
;	    periodiclist=periodiclist, pixel=pixel, plot=plot, $
;	    _extra=_extra] )
; INPUTS:
;	tsd		TSD object reference.
;	cal_const_file	Calibration constants file;
;			not yet implemented
;	pixel		operate on these pixels only
;			default: all
;	plot		make plots
;	_extra		for plotting.
; OUTPUTS:
;	tsd		TSD object reference. Deglitched TSD.
;	periodiclist	make list of the results, 1 line per detector
;			How to manage the summaries is under considerations.
;RETURN
;	!SUCCEED if function ended successfully.
;	!FAIL in failure.
; RESTRICTIONS:
;	Only works for SW. Needs to be tested more. Run before the deglitchers.
; PROCEDURE:
;
; EXAMPLE:
;	See README
; REFERENCE:
;	Baba et al. 2002 Nov. 21, ``ASTRO-F/FIS Survey Data Structure"
;
; MODIFICATION HISTORY:
;	Oct  2006   DK	Creation
;-

FUNCTION GB_periodic_noise, tsd, Calib_Const_File = ccf, $
	    periodiclist=periodiclist, pixel=pixel, plot=plot, $
	    _extra=_extra

;   ON_ERROR 2

;;; Starting GB_periodic_noise ;;;
    lmessage, 7,'Executing GB_periodic_noise...'

    if ( n_elements( periodiclist ) GT 0 ) then begin
	if ( size( periodiclist, /type ) EQ 7 ) then listfile = periodiclist $
	else begin
	    infile = tsd->get_filename()
	    str_len = strpos( infile, ".fits" )
	    listfile = strmid( infile, 0, str_len ) + '_periodic_list.dat'
	endelse
	openw, glunit, listfile, /get_lun
    endif

    pltype = size( plot, /type )
    if ( pltype GT 0 ) then begin
	do_plot = 1
	if ( pltype EQ 7 ) then start_hc, plot $
	else do_plot = plot
    endif

;    if ( n_elements( ccf ) EQ 0 ) then ccf = 'PERN_'
;    frincal = ccf_read_ccf( ccf, tsd, error=err )
;    if ( err EQ !FAIL ) then $
;	message, "No CCF file found for periodic noise removal", /continue

    type = tsd->get_datatype()
    if ( type EQ 'FIS_SW' ) then begin
	detstep = 20
	if ( n_elements( pixel ) EQ 0 ) then pixel = indgen( 100 ) + 1
    endif
    if ( type EQ 'FIS_LW' ) then begin
	detstep = 15
	if ( n_elements( pixel ) EQ 0 ) then pixel = indgen( 75 ) + 1
    endif

    npixel = n_elements( pixel )

    flux = perinoise( tsd, pixel, noisemodel=noisemodel, tukscale=tukscale, $
	    detstep=detstep, unit=glunit, flags=flags, $
	    plot=do_plot, status=status, _extra=_extra )

    if ( status EQ 0 ) then begin
	status = tsd->set_value_tsd( 'fis_obs', 'flux', flux )
	status = tsd->set_pix_flag_tsd( 'fis_obs', 'no_peri_corr', flags )
    endif

    if ( n_elements( periodiclist ) GT 0 ) then $
	free_lun, glunit
    if ( pltype EQ 7 ) then stop_hc

;;; Write HISTORY in header ;;;

    lmessage, 7, 'Writing HISTORY',/CONTINUE
    hist_message = ' gb_periodic_noise.pro,v 0.0 done. CCF= '

    IF KEYWORD_SET(ccf) NE 1 THEN BEGIN
	hist_message = hist_message + 'NONE'
    ENDIF ELSE BEGIN
	hist_message = hist_message + ccf
    ENDELSE

    dummy = tsd_fxaddhist(tsd, hist_message)
    IF dummy NE 0 THEN $
	MESSAGE,'Unable to write HISTORY',/CONTINUE

;    print, systime(1) - syst

    return, status

END


