; $Id: gb_bayes_glitch.pro,v 1.5 2006/10/06 08:51:02 do Exp $
;
; Copyright (c) 2003, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;	gb_bayes_glitch.pro
; PURPOSE:
;	Apply a Bayesian GaussianProcess glitch finder to one pixel
; CATEGORY:
;	Function. Library for the procedures in ASTRO-F pipeline.
; CALLING SEQUENCE:
;	flag = gb_bayes_glitch( flux, inflag, status, glitcal, qual=qual, $
;		 unit=unit, plot=plot, pixel=pixel, xr=xr, yr=yr, seed=seed )
;
; INPUTS:
;	flux		fluxarray of one pixel
;	inflag		input flags relevant to deglitching
;	status		of the instument (cannot deglitch over diff statuses)
;	glitcal 	calibration structure
;	qual		quality flag (2 bits)
;			0: OK: no (indication of a) glitch found
;			1: doubtfull glitch or (non-glitch) or
;			   could not determine (too few points)
;			2: poor quality glitch
;			3: good quality glitch
;	unit		to write the glitch list to
;	plot		make plots
;	pixel		pixel number (just for reporting)
;	xr		xrange in plot
;	yr		yrange in plot
;	seed		for inserting random spikes into flux
;			default: none
;
; OUTPUTS:
; RETURN:
;	flag		glitch flags are set
; RESTRICTIONS:
;
; PROCEDURE:
;
; EXAMPLE:
;
; REFERENCE:
;
; REFERENC00
;	This function uses append.pro and edgemedian.pro in
;	the library of Green box.
; MODIFICATION HISTORY:
;	Dec. 2003		Written by Do Kester.
;	June 2004		Update
;	June 2005 DK further development of gb_gauss_glitch
;-

; ======================================================================
; right shift with zero padding

function rshift, indx

indx[0] = 0
return, shift( indx, -1 )

end

; ======================================================================
;; Set up the covariance matrix

function make_corrmat, time, radius, amp

npoint = n_elements( time )

cov = dblarr( npoint, npoint )
for k = 0, npoint-1 do begin
    cov[k,*] = amp * exp( -0.5 * ( ( time[k] - time ) / radius )^2 )
endfor

return, cov
end

; ======================================================================

pro make_cm, time, rad, th1, npt

common cormat, cov, ones, npts

npoint = n_elements( time )

cov = make_corrmat( time, rad, th1 )

ones = intarr( npt ) + 1
npts = intarr( npt ) + npoint

return
end

; ======================================================================

function get_cm, time

common cormat, cov, ones, npts

return, cov[time ## ones + npts ## time]

end

; ======================================================================
function solvegp, flux, mat, noise, fit, nspike, logprior, lumat=lumat, $
		 clean=clean

common fupgp, npt, cov, diag, sol, mean, s2

common precalc, logsqrtpi, logdof

s2 = noise * noise
npt = n_elements( flux )

mean = median( flux )
vec = ( flux - mean )			; * th1
;;sol = lusol( cov, index, vec )		; backsubstitution
sol = cholsol( cov, diag, vec )

fit = mat # sol + mean

;chisq = total( ( flux - fit ) ^ 2 )
chisq = total( ( flux - fit ) ^ 2 ) / s2

dof = npt - nspike
evid = logsqrtpi + logdof[dof] - nspike * logprior + $
       npt - 0.5 * dof * ( 1 + alog( 2 * !PI * chisq / dof ) )

if ( n_elements( lumat ) GT 0 ) then clean = lumat # sol + mean

return, evid

end

; ======================================================================

; returns evidence = log( Pr )
function fitgp, flux, mat, noise, fit, nspike, logprior, lumat=lumat, $
		 clean=clean

common fupgp, npt, cov, diag, sol, mean, s2

cov = mat
s2 = noise * noise

;; Add noise to the trace
npt = n_elements( flux )
q = indgen( npt )
cov[q,q] = cov[q,q] + noise

;; LU decomposition
;;ludc, cov, index, interchanges=parity
diag = dblarr( npt )
choldc, cov, diag

return, solvegp( flux, mat, noise, fit, nspike, logprior, lumat=lumat, $
		 clean=clean )
end

; ======================================================================
pro makestdcov, mat, noise

common stdchol, cholcov, stddiag

cholcov = mat
s2 = noise * noise

;; Add noise to the trace
npt = n_elements( mat[0,*] )
q = indgen( npt )
cholcov[q,q] += noise

;; Cholesky LU decomposition
stddiag = dblarr( npt )
choldc, cholcov, stddiag

return
end

; ======================================================================
function stdgp, flux, mat, noise, fit, nspike, logprior, lumat=lumat, $
		 clean=clean

common fupgp, npt, cov, diag, sol, mean, s2
common stdchol, cholcov, stddiag

cov = cholcov
diag = stddiag

return, solvegp( flux, mat, noise, fit, nspike, logprior, lumat=lumat, $
		 clean=clean )
end

; =================================================================
function updategp, flux,  mat, noise, fit, nspike, logprior, lumat=lumat, $
		 clean=clean, loc=loc, spike=spike

common fupgp, npt, cov, diag, sol, mean, s2

common precalc, logsqrtpi, logdof

uv = dblarr( npt )
uv[loc] = spike

zed = cholsol( cov, diag, uv )
;zed = lusol( cov, index, uv )

sup = sol - zed * sol[loc] / ( 1 + zed[loc] )

fit = mat # sup + mean

;chisq = total( ( flux - fit ) ^ 2 )
chisq = total( ( flux - fit ) ^ 2 ) / s2

dof = npt - nspike
evid = logsqrtpi + logdof[dof] - nspike * logprior + $
       npt - 0.5 * dof * ( 1 + alog( 2 * !PI * chisq / dof ) )

if ( n_elements( lumat ) GT 0 ) then clean = lumat # sup + mean

return, evid
end

; ====================================================================
; ====================================================================

function gb_bayes_glitch, flux0, inflag, status, glitcal, qual=qual, $
	unit=unit, plot=plot, pixel=pixel, xr=xr, yr=yr, dy=dy, seed=seed

COMPILE_OPT IDL2

; set common constant and lookup table
common precalc, logsqrtpi, logdof
logsqrtpi = 0.5 * alog( !PI )
logdof = alog( findgen( glitcal.npoint + 1 ) )

nflux0 = n_elements( flux0 )
time0 = lindgen( nflux0 )
bval = where( inflag EQ 0, nsam )

sect = status - shift( status, 1 )
sect[0] = 0
bsec = [where( sect NE 0, nbs ), nflux0]    ;; first element of each section

;; check whether there are samples at all, if not return empty flag array
outflag = bytarr( nflux0 )
if ( nsam EQ 0 ) then return, outflag

flux1 = flux0[bval]
time = time0[bval]
;noise = median( abs( flux - shift( flux, 1 ) ) )

;; Add spikes (random in place and size) to the data stream
if ( n_elements( seed ) GT 0 ) then begin
    nspi = 100
    bran = long( randomu( seed, nspi ) * nsam )
    spik = randomn( seed, nspi ) * 0.0001
    flux1[bran] += spik
endif

rad	= glitcal.radius
th1	= glitcal.amplit
th2	= glitcal.offset
;th3	= glitcal.noise
th4	= glitcal.slope
npoint	= glitcal.npoint     ; size of environment
lprior	= alog( glitcal.prior )

; The prior should maybe be as low as 0.5 as the level at which saturation
; sets in. I have to time to investigate this further. Could get more glitches.
; lprior = alog( 1.e3 )

spike_c = glitcal.spike
qlevel	= glitcal.qlevel

;print, nsam, npoint, spike_c
th3 = median( abs( flux1 - shift( flux1, 1 ) ) ) * 2 > glitcal.noise
;print, pixel, rad, th1, th2, th3, th4, npoint, lprior, glitcal.noise
;print, flux0[0:5]

mid = npoint / 2

;; Plotting the stuff
if ( keyword_set( plot ) ) then begin
    nbxr = 0
    if ( n_elements( xr ) EQ 2 ) then $
	bxr = where( bval GE xr[0] AND bval LE xr[1], nbxr )
    if ( nbxr EQ 0 ) then bxr = bval
    if ( n_elements( yr ) EQ 0 ) then begin
	yr = [min( flux1[bxr] ), max( flux1[bxr] )]
    endif
     plot, flux0, ps=1, xst=3, yst=3, xr=xr, yr=yr, $
	    position=[0.1,0.3,0.9,0.9], title=string( "Pixel Nr ", pixel ), $
	    xtickname=replicate( ' ', 10 )

; overplot different statuses at lower axis
    unstat = status[ uniq( status, sort( status ) ) ]
    for ku = 0, n_elements( unstat ) -1 do begin
	us = unstat[ku]
	bu = where( status EQ us, nbu )
	oplot, bu, fltarr( nbu ) + !y.crange[0], ps=1, color=ku+1
    endfor
; overplot resets and dead pixels
    bb = where( inflag GT 0, nbb )
    if ( nbb GT 0 ) then begin
	oplot, bb, flux0[bb], ps=4, color=8
	oplot, bb, fltarr( nbb ) + !y.crange[0], ps=4, color=8
    endif

    if ( n_elements( seed ) GT 0 ) then $
	oplot, [bval[bran]], [flux1[bran]], ps=4, color=5

;    if ( nb GT 0 ) then oplot, [bval[b]], [flux1[b]], ps=4, color=2
endif

q = indgen( npoint )
nsam = nflux0
smfit = fltarr( nsam )
backg = fltarr( nsam )
dbell = fltarr( nsam )
glist = intarr( nsam )
tlist = intarr( nsam )

ntrans = 0L
cpu = 0.0
;looptime = systime(1)
;cormtime = 0

make_cm, indgen( 2*npoint ), rad, th1, npoint
makestdcov, get_cm( q ), th3

to_dbell = 10 / alog( 10 )
si = intarr( npoint )		  ; spike index
b1 = -1
simple = 0
toolon = 0
for ks = 0, nbs do begin
  b0 = b1
  b1 = bsec[ks]

;oplot, [b0,b1], [0.0001,0.0001]*(ks+1)

  bv = where( bval GE b0 AND bval LT b1, nsam )
  if ( nsam LE npoint ) then begin
    if ( nsam GT 0 ) then begin
      tt = time0[bval[bv]]
      outflag[tt] = 9		    ; impossible to determine
      dbell[tt] = 1		    ; to signal dubiousness
      smfit[tt] = flux1[bv]
      backg[tt] = flux1[bv]
    endif
    continue			    ; TBD set quality flags
  endif

;oplot, bval[bv], fltarr(nsam)+0.0001*(ks+1)+0.004, color=2

  flux = flux1[bv]
;oplot, bval[bv], flux, color=2, ps=1

  bv = bval[bv]

  time = time0[bv]

;oplot, time, flux0[bv], color=3, ps=1

  for k = 0, nsam - npoint do begin
    tt = time[q+k]
    fl = flux[q+k]
;oplot, tt, fl, ps=1, color=2

;cpu1 = systime(1)
;    cov0 = make_corrmat( tt, rad, th1 )
    tt0 = tt - tt[0]
    if ( tt0[npoint-1] LT 2 * npoint ) then cov0 = get_cm( tt0 ) $
    else begin
	cov0 = make_corrmat( tt, rad, th1 )
	toolon++
    endelse

;cormtime += ( systime(1) - cpu1 )
    mat = cov0
; calculate the basic evidence
;    cpu1 = systime(1)
    if ( tt0[npoint-1] GE npoint OR ispresent( si EQ 1, bs, ns ) ) then begin
	if ( ns GT 0 ) then cov0[bs,bs] += spike_c
	evidence = fitgp( fl, cov0, th3, fit0, ns, lprior, lumat=mat, $
			  clean=clean )
	ntrans++
    endif else begin
	evidence = stdgp( fl, cov0, th3, fit0, ns, lprior, lumat=mat, $
			  clean=clean )
	simple++
    endelse


;    cpu += ( systime(1) - cpu1 )

    bestfit = fit0
    midcl = clean[mid]
    midbg = fit0[mid]

; avoid edge points when determining the best location of a glitch
    mx = max( abs( (fl - fit0)[1:npoint-2] ), kmax )
    kmax += 1		    ; add 1 to correct for edge avoidance
; add a spike in the position of the max deviation
    cov1 = cov0
    cov1[kmax,kmax] += spike_c
    evi1 = updategp( fl, cov1, th3, fit1, ns+1, lprior, lumat=mat, $
		     clean=clean, loc=kmax, spike=spike_c )

;    evi1 = fitgp( fl, cov1, th3, fit1, ns+1, lprior, lumat=mat, clean=clean )

    if ( evi1 GT evidence ) then begin
	dbell[tt[kmax]] = ( evi1 - evidence ) * to_dbell
	si[kmax] = 1
	evidence = evi1
	bestfit = fit1
	midcl = clean[mid]
	midbg = fit1[mid]
	cov0 = cov1
    endif

    if ( k EQ 0 ) then begin
	smfit[time[0]+q] = clean
	backg[time[0]+q] = fit0
	glist[time[0]+q] = si		; catch glitches in first semi-window
    endif

;try to kick out the midpoint spike
    if ( si[mid] EQ 1 ) then begin
	cov1 = mat
	si[mid] = 0
	if( ispresent( si EQ 1, bs, ns ) ) then cov1[bs,bs] += spike_c

;    cpu1 = systime(1)
	evi1 = fitgp( fl, cov1, th3, fit1, ns, lprior, lumat=mat, clean=clean )
	ntrans++
;    cpu += ( systime(1) - cpu1 )

	if ( evi1 GT evidence ) then begin
	    evidence = evi1
	    bestfit = fit1
	    midcl = clean[mid]
	    midbg = fit1[mid]
	    cov0 = cov1
	endif else si[mid] = 1			; restore
    endif

    kk = tt[mid]
    smfit[kk] = midcl
    backg[kk] = midbg
    glist[kk] = si[mid]

    si = rshift( si )
  endfor

  glist[tt[mid:*]] = si[mid-1:npoint-2] 	; correct for rshift
  smfit[tt[mid:*]] = clean[mid:*]
  backg[tt[mid:*]] = bestfit[mid:*]
endfor

flux = flux0[bval]		;; restore
time = time0[bval]
smfit = smfit[bval]
backg = backg[bval]
glist = glist[bval]
dbell = dbell[bval]
;print, 'NTRANS = ', ntrans, simple, toolon, cpu, systime(1)-looptime, cormtime

if ( n_elements( qual ) GT 0 ) then begin
    if ( ispresent( dbell GT 0, bq ) ) then qual[bq] = 1
    if ( ispresent( dbell GT qlevel[0] AND glist EQ 1, bq ) ) then qual[bq] = 2
    if ( ispresent( dbell GT qlevel[1] AND glist EQ 1, bq ) ) then qual[bq] = 3
endif

out = flux - smfit

if ( keyword_set( plot ) ) then begin
;    maxdb = max( dbell[bxr] )
;    print, maxdb
    oplot, time, smfit, color=3
;    oplot, time, dbell * yr[1] / maxdb, color=6
    if ( ispresent( glist EQ 1, bl ) ) then $
	oplot, bval[bl], flux[bl], color=2, ps=1

    if ( n_elements( dy ) EQ 0 ) then dy = max( abs( out[bxr] ) )
    if ( n_elements( dy ) EQ 1 ) then dy = [-dy,dy]
;    res = flux - backg
;    plot, bval, res, xst=3, yst=3, xr=xr, yr=dy, $
    res = flux - backg
    plot, time, res, xst=3, yst=3, xr=xr, yr=dy, yminor=0, yticks=2, $
	position=[0.1,0.1,0.9,0.3], /noerase, xtitle="sample index"

    if ( ispresent( glist EQ 1, bl ) ) then $
	oplot, bval[bl], out[bl], color=2, ps=1
    if ( n_elements( qual ) GT 0 ) then $
	oplot, time, 0.1 * dy[1] * qual - dy[1], color=6, ps=3
    if ( ispresent( outflag EQ 9, bl, nbl ) ) then $
	oplot, time0[bl], fltarr( nbl ) - 0.9 * dy[1], color=7, ps=3
endif

outpos = glist EQ 1 AND out GT 0		    ; positive outliers
outneg = glist EQ 1 AND out LT 0		    ; negative outliers
conseq = ( shift( outpos, 1 ) AND outpos ) OR $     ; + or - tails
	 ( shift( outneg, 1 ) AND outneg )

if ( ispresent( outpos, qp ) ) then begin
    outflag[bval[qp]] = 1		    ; positive spike
    if ( keyword_set( plot ) ) then $
	oplot, bval[qp], ( out[qp] < dy[1] ), ps=2, color=5
endif

if ( ispresent( outneg, qp ) ) then begin
    outflag[bval[qp]] = 2		    ; negative spike
    if ( keyword_set( plot ) ) then $
	oplot, bval[qp], ( out[qp] > dy[0] ), ps=2, color=6
endif

if ( ispresent( outneg AND shift( outpos, 1 ), qp, nqp )) then begin
    qp = reform( [qp,qp-1], 2*nqp ) > 0
    outflag[bval[qp]] = 3		    ; pos-neg spike
    if ( keyword_set( plot ) ) then $
	oplot, bval[qp], ( out[qp] < dy[1] ), ps=2, color=7
endif

if ( ispresent( conseq, qp, nqp )) then begin
    outflag[bval[qp]] = 5		    ; tail
    if ( keyword_set( plot ) ) then $
	oplot, bval[qp], ( out[qp] < dy[1] ), ps=2, color=8
endif

;; Print to a logfile (at unit)
if ( n_elements( unit ) GT 0 ) then begin
    b = where( outflag[bval] GT 0 AND outflag[bval] LT 6, n0 )
;    printf, unit, 'Deglitching pixel number ', pixel
    for k = 0, n0-1 do begin
	bk = b[k]
	printf, unit, pixel, k+1, bval[bk], outflag[bval[bk]], $
		1000000*out[bk], dbell[bk], qual[bk], $
		format='(i3,i6,i8,i4,f12.1,f10.1,i4)'
    endfor
endif

return, outflag

end

