; $Id: gb_gauss_glitch.pro,v 1.5 2006/01/18 06:57:00 baba Exp $
;
; Copyright (c) 2003, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;	gb_gauss_glitch.pro
; PURPOSE:
;	Apply the GaussianProcess glitch finder to one pixel
; CATEGORY:
;	Function. Library for the procedures in ASTRO-F pipeline.
; CALLING SEQUENCE:
;	flag = gb_gauss_glitch( flux, inflag, glitcal, $
;		 unit=unit, plot=plot, pixel=pixel, xr=xr, yr=yr, seed=seed )
;
; INPUTS:
;	flux		fluxarray of one pixel
;	inflag		input flags relevant to deglitching
;	glitcal 	calibration structure
;	unit		to write the glitch list to
;	plot		make plots
;	pixel		pixel number (just for reporting)
;	xr		xrange in plot
;	yr		yrange in plot
;	seed		for inserting random spikes into flux
;			default: none
; OUTPUTS:
; RETURN:
;	flag		glitch flags are set
; RESTRICTIONS:
;
; PROCEDURE:
;
; EXAMPLE:
;
; REFERENCE:
;
; REFERENC00
;	This function uses append.pro and edgemedian.pro in
;	the library of Green box.
; MODIFICATION HISTORY:
;	Dec. 2003		Written by Do Kester.
;	June 2004		Update
;-

function gb_gauss_glitch, flux0, inflag, glitcal, $
		unit=unit, plot=plot, pixel=pixel, xr=xr, yr=yr, seed=seed

COMPILE_OPT IDL2

theta	= glitcal.theta 	    ; gaussian parameter set
npoint	= glitcal.npoint	    ; size of environment
iter	= glitcal.iter		    ; maximum nr of iterations
cutoff	= glitcal.cutoff	    ; S/N
boxlen	= glitcal.boxlen	    ; smoothing length of bandwidth
neborin = glitcal.neborin	    ; influence of neightbouring samples
source	= glitcal.source	    ; influence of point sources

dt  = 2.0 / npoint
medcut = cutoff / 0.6			    ; transform to median

rad = theta[0]
th1 = theta[1]
th2 = theta[2]
th3 = theta[3]
th4 = theta[4]

;print, theta

;; Set up the covariance matrix
mid = npoint / 2
tt  = ( findgen( npoint ) - mid ) * dt
rad = rad * dt
cov = dblarr( npoint, npoint )
mat = cov
for k = 0, npoint-1 do begin
    cov[k,*] = th1 * exp( -0.5 * ( ( tt[k] - tt ) / rad )^2 ) + th2
    cov[k,*] = cov[k,*] + th4 * tt[k] * tt
    mat[k,*] = cov[k,*]
;    print, cov[k,*], format='(20f10.6)'
endfor

;; Add noise to the trace
q = indgen( npoint )
cov[q,q] = cov[q,q] + th3
;cov[mid,mid] = cov[mid,mid] + th3

;; LU decomposition
ludc, cov, index, interchanges=parity
bval = where( inflag EQ 0, nsam )

;; check whether there are samples at all, if not return empty flag array
if ( nsam EQ 0 ) then return, bytarr( n_elements( flux0 ) )

flux = flux0[bval]

;; Add spikes (random in place and size) to the data stream
if ( n_elements( seed ) GT 0 ) then begin
    nspi = 100
;    nspi = 2
    bran = long( randomu( seed, nspi ) * nsam )
    spik = randomn( seed, nspi ) * 0.0001
    flux[bran] += spik
endif


fluc = flux			; temp version of samples for calculations

;; Iterate to get rid off the influence of the spikes on the "smooth" steam
;iter = 1
for n = 0, iter do begin
  bn = lindgen( nsam )
  bb = where( bn LT nsam - npoint, nb )
;print, n, nb
  smo = fltarr( nsam )
  q = indgen( npoint )
  for kk = 0L, nb - 1 do begin		; iterate over the samples
    k = bn[kk]
    fl = fluc[q+k]
    mean = median( fl )
    mean = total( fl ) / npoint
    vec = ( fl - mean ) * th1
    sol = lusol( cov, index, vec )		; backsubstitution

    fit = mat # sol + mean
    smo[mid+k] = fit[mid]			; take midpoint as smooth
;if ( kk EQ 1550 ) then begin
;    print, "flux ", fl
;    print, "fit  ", fit
;endif
  endfor

;; fix the edges
  smo[0:mid] = smo[mid]
  smo[nsam-mid-1:nsam-1] = smo[nsam-npoint-1+mid]

;; Calculate residuals and a bandwidth of 5 sigma
  res = flux - smo
  absres = abs( res )
  mr5 = smooth( edgemedian( absres, boxlen ), boxlen, /edge_truncate ) * $
		medcut
  mr2 = fltarr( nsam )
  for k = 3, 1, -1 do begin
    mr2 += shift( absres, k ) > shift( absres, -k )	; max of neighbours
    mr2 *= neborin
  endfor
  mr3 = smooth( edgemedian( ( ( flux - shift( flux, 1 ) ) > 0 ) > $
     ( ( flux - shift( flux, -1 ) ) > 0 ), 3 ), 3, /edge_truncate ) * source
  mr2 = mr3 > mr2

;; Find samples outside the bandwidth
  rem = absres GT mr5
  rem = rem OR shift( rem, 1 ) OR shift( rem, -1 )  ; add previous and next pix

;; Identify which of a contiguous series is the glitch (a bit of magick)
  bn = where( rem, nb )
  if ( nb GT 0 ) then begin
    b1 = where( bn - shift( bn, 1 ) GT 1, nb1 )     ; find first of each series
    if ( nb1 GT 0 ) then b1 = [ 0L, b1, nb ] else b1 = [0, nb]
    nb1 = nb1 + 1				    ; add first and last
    for k = 1L, nb1 do begin			    ; process per serie
	st = bn[b1[k-1]]
	la = bn[b1[k]-1]
	bsl = indgen( la - st + 1 ) + st
	mx = max( absres[bsl], maxat )
	maxat += st
	b4 = append( b4, maxat )		    ; max is always glitch
;	bsl = bsl[where( bsl NE maxat )]
	bs = where( absres[bsl] GT mr5[bsl] AND absres[bsl] GT mr2[bsl], nbs )
	if ( nbs GT 0 ) then b4 = append( b4, bsl[bs] )

    endfor
    b = temporary( b4 )
  endif else break
;print, b

;; replace the glitches with the smooth version
  fluc = flux
  if ( nb GT 0 ) then fluc[b] = smo[b]
endfor

;; Plotting the stuff
if ( keyword_set( plot ) ) then begin
    nbxr = 0
    if ( n_elements( xr ) EQ 2 ) then $
	bxr = where( bval GE xr[0] AND bval LE xr[1], nbxr )
    if ( nbxr EQ 0 ) then bxr = bval
    if ( n_elements( yr ) EQ 0 ) then begin
	yr = [min( smo[bxr] - 4*mr5[bxr] ), max( smo[bxr] + 4*mr5[bxr] )]
    endif
;   plot all points
    plot, flux0, ps=1, xst=3, yst=3, xr=xr, yr=yr, $
	    position=[0.1,0.3,0.9,0.9], title=string( "Pixel Nr ", pixel ), $
	    xtickname=replicate( ' ', 10 )
    bb = where( inflag GT 0, nbb )
;   overplot flagged points with diamonds in ocher
    if ( nbb GT 0 ) then oplot, bb, flux0[bb], ps=4, color=8

;   overplot randomly placed extra points (diamond yellow)
    if ( n_elements( seed ) GT 0 ) then $
	oplot, [bval[bran]], [flux[bran]], ps=4, color=5

;    if ( nb GT 0 ) then oplot, [bval[b]], [flux[b]], ps=4, color=2
endif

b = where( abs( res ) GT mr5, nb )
b = where( abs( res ) GT mr5 AND abs( res ) GT mr2, nb )

if ( keyword_set( plot ) ) then begin
    oplot, bval, smo, color=3
;    oplot, bval, smo+mr2, color=4
;    oplot, bval, smo-mr2, color=4
    if ( nb GT 0 ) then oplot, bval[b], ( flux[b] < yr[1] ), ps=1, color=2

;    maxr5 = max( mr5[bxr] ) > max( mr2[bxr] )
    maxr5 = max( mr5[bxr] )
    dy = [-maxr5,maxr5] * 2

    plot, bval, res, xst=3, yst=3, xr=xr, yr=dy, $
	position=[0.1,0.1,0.9,0.3], /noerase, xtitle="sample index"

    oplot, bval, mr5, color=3
    oplot, bval, -mr5, color=3
    oplot, bval, mr2, color=4
    oplot, bval, -mr2, color=4

endif

;; Set up a flag array
flag = bytarr( n_elements( flux0 ) )
if ( nb GT 0 ) then begin
    if ( keyword_set( plot ) ) then oplot, bval[b], res[b], ps=2, color=2

    qp = where( res[b] GT 0, nqp )		; positive glitches
    if ( nqp GT 0 ) then begin
	flag[bval[b[qp]]] = 1
	if ( keyword_set( plot ) ) then $
	    oplot, bval[b[qp]], ( res[b[qp]] < dy[1] ), ps=2, color=5
    endif

    qn = where( res[b] LT 0, nqn )		; negative glitches
    if ( nqn GT 0 ) then begin
	flag[bval[b[qn]]] = 2
	if ( keyword_set( plot ) ) then $
	    oplot, bval[b[qn]], ( res[b[qn]] > dy[0] ), ps=2, color=6
    endif

    qq = where( res[b] GT 0 AND shift( b, -1 ) EQ b+1 AND res[b+1] LT 0, nqq )
    if ( nqq GT 0 ) then begin			; tails
	qq = bval[b[qq]]
	while ( nqq GT 0 ) do begin
	    flag[qq] = 3
	    qq = where( flag EQ 2 AND shift( flag, 1 ) EQ 3, nqq )
	endwhile

	qq = where( flag[bval] EQ 3 )
	if ( keyword_set( plot ) ) then $
	    oplot, bval[qq], ( ( res[qq] < dy[1] ) > dy[0] ), ps=2, color=7
    endif

endif

;; Print to a logfile (at unit)
b = where( flag[bval] NE 0, n0 )
if ( n_elements( unit ) GT 0 ) then begin
;    printf, unit, 'Deglitching pixel number ', pixel
    if( n0 GT 0 ) then begin
	for k = 0, n0-1 do begin
	    bk = b[k]
	    printf, unit, pixel, k+1, bval[bk], flag[bval[bk]], $
		1000000*res[bk], 3 * abs(res[bk]) / mr5[bk], $
		format='(i3,i6,i8,i4,f12.1,f10.1)'
	endfor
    endif
endif

;b = where( flag EQ 1, n1 )
;b = where( flag EQ 2, n2 )
;b = where( flag EQ 3, n3 )
;print, "Iterations ", n, n0, n1, n2, n3

if ( n_elements( seed ) GT 0 ) then begin
    bvr = bval[bran]
    bpp = where( flag[bvr] NE 0, npp )
    bpn = where( flag[bvr] EQ 0, npn )

    qqq = intarr( n_elements( flux0 ) )
    qqq[bvr] = 1
    bvn = where( qqq EQ 0, nbc )
    bnp = where( flag[bvn] NE 0, nnp )
    bnn = where( flag[bvn] EQ 0, nnn )

;    print, npp, npn
;    print, nnp, nnn
endif

return, flag

end

