;
; Copyright (c) 2002, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;	mmt_deglitcher.pro
; PURPOSE:
;       Deglitching engine to flag (and remove) glitches in TSD
; CATEGORY:
;	Green Box procedure in library in ASTRO-F pipeline.
; CALLING SEQUENCE:
;       mmt_deglitcher, data, 
; INPUTS:
;       timeline     input data
;
; OUTPUTS:
;       data         output data 
;
; RETURN
;	!SUCCEED if function ended successfully.
;	!FAIL in failure.
; RESTRICTIONS:
;	
; PROCEDURE:
;	1. This function will be provided by the European team.
;	2. The after effects are corrected with the transient correction.
; EXAMPLE:
;
; REFERENCE:
;
; MODIFICATION HISTORY:
; 	Mar. 2004		Written by T.Takagi
;    
;-

pro mmt_deglitcher, data_orig, data_out, sigma, $
                    sigma_j = sig_j, $
                    threshold=threshold,scale=scale,$
                    pspikes=pspikes, nspikes=nspikes, dippers=dippers, $
                    faders=faders, sources=sources, $ 
                    debug=debug, glitch_tsd=glitch_tsd, $
                    glitch_measures=glitch_measures,$
                    ch=ch, subid=subid
             ;; keywords for debug


if (n_elements(debug) eq 0) then debug = 0

time = dblarr(10)
time[0] = systime(/second)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; MMT set up
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
if (n_elements(scale) eq 0) then scale=2L^[0,1,2,3,4,5,6,7,8,9] + 1

;; make sure that the input is one dimension array
data = data_orig  ; array 'data' will be modified and output 
;;- gaussian noise 
;data = randomu(seed, n_elements(data), /normal)

;- flag array
flg_dip = bytarr(n_elements(data))
flg_fad = bytarr(n_elements(data))
flg_src = bytarr(n_elements(data))

time[1] = systime(/second)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; MMT 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
mmt, data, wave, scale=scale

time[2] = systime(/second)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Prepare noise level at each scale
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
sigma_j = sig_j*sigma
time[3] = systime(/second)
;print, sigma_j

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Object detection and labelling 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
mmt_object_id, wave, sigma_j, threshold, mask, proj, struct, nstrs

;- if there is no structure, return to the gb process

if (nstrs eq 0 or max(abs(proj)) le 1) then begin 
    LMESSAGE,7,'No structures',/CONTINUE    
    faders = 0
    dippers = 0
    sources = 0
    glitch_tsd = 0
    glitch_measures = 0
    return
endif 


time[4] = systime(/second)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Determine proj_limit 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
proj_limit = max(proj) < max(-proj)
if (proj_limit le 1) then proj_limit = 2
               ;; If there is readout with projected mask grater than 
               ;; proj_limit, they are candidates of glitchs. 
               ;;    -> this is solved with considering negative proj

if (debug eq 1) then begin
    LMESSAGE,7, 'Project Mask Limit = '+strcompress(proj_limit),/CONTINUE
endif

time[5] = systime(/second)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Reconstuct TSD at large projected mask
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

isBig = bytarr(nstrs+1)
usedBig = bytarr(nstrs+1)
part_wave=fltarr(n_elements(data), n_elements(scale))    
big_tsd = fltarr(n_elements(data),2) 
 ; tsd for each positive and negative proj. 
 ; Sometimes, positive proj is found for dippers, and vice versa
 ; Detemination of fader or dipper will be done below

id_read = where(abs(proj) ge proj_limit)
for i=0,n_elements(id_read)-1 do begin
    arr=struct[id_read[i],*]
    isBig[arr]=1

    id_big = where(isBig eq 1,cnt)
    
    for ibig=0,n_elements(id_big)-1 do begin 
        str_id = id_big[ibig]
        if (str_id ne 0 and usedBig[str_id] ne 1) then begin 
            id = where(struct eq str_id)
            part_wave[id] = wave[id] 
            mmt_inverse, part_wave, ptsd, scale=scale

            if (proj[id_read[i]] lt 0) then begin 
                big_tsd[*,0] = big_tsd[*,0] + ptsd
            endif else begin 
                big_tsd[*,1] = big_tsd[*,1] + ptsd
            endelse

            part_wave[id]=0
            usedBig[str_id]=1
        endif
    endfor
    isBig[arr]=0
endfor

time[6] = systime(/second)


time[7] = systime(/second)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Cleaning of dipper and fader TSD
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;- fader and dipper TSD may include possible sources 

;; PARAMETERS !! ;;; 
sep_limit = 10 ;; Individual objects in dipper_ and fader_tsd are 
               ;; assumed to be separated by sep_limit [readout]

sep_threshold = sigma*0.1  ;; threshold to separate signals

fraction_at_width = 0.2 ;; fraction of value at which the width is measured

ratio_limit = 3. ;; limit for width ratio 

source_size = 30 ;; if a signal duration is shorter than this, 
                 ;;  the signal is regarded as source

;; PARAMETERS END ;;; 

;- dippers first 
tsd1 = reform(big_tsd[*,0])
mmt_check_glitch_tsd, tsd1,proj, sigma, $
  sep_threshold, sep_limit, $
  ratio_limit, fraction_at_width, source_size,$
  glitch_measures=gm1, flg_dip, flg_fad, flg_src, pspikes

;- fader 
tsd2 = reform(big_tsd[*,1])
mmt_check_glitch_tsd, tsd2,proj, sigma,  $
  sep_threshold, sep_limit, $
  ratio_limit, fraction_at_width, source_size, $
  glitch_measures=gm2, flg_dip, flg_fad, flg_src,pspikes

glitch_measures = mmt_define_glitch_measures(0,gm1=gm1,gm2=gm2)

time[8] = systime(/second)


;;;
;;; Check computation time
;;;
elapsed_time = time[1:8]-time[0:7]
;print, elapsed_time
;print, 'total time = ',time[8]-time[0]
;stop 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Subtraction of glitch TSD
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
dipper_tsd = dblarr(n_elements(data))
fader_tsd = dblarr(n_elements(data))

id_dip = where(flg_dip eq 1, cnt1)
id_fad = where(flg_fad eq 1, cnt2)
if (cnt1 ne 0) then begin     
    dipper_tsd[id_dip] = tsd1[id_dip]+tsd2[id_dip]
endif
if (cnt2 ne 0) then begin
    fader_tsd[id_fad] = tsd1[id_fad]+tsd2[id_fad]
endif

glitch_tsd = dipper_tsd + fader_tsd 
data_out = data - glitch_tsd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; flag
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

faders = where(flg_fad eq 1,cnt)
dippers = where(flg_dip eq 1,cnt)
sources = where(flg_src eq 1,n_src)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; debug plotting
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;; debug ;;;;
if (debug eq 1) then begin 

    dum = where(glitch_tsd ne 0., cnt)
    cnt = 1
    if (cnt ne 0) then begin 
        !p.multi=[0,1,4]
        af_sky_stats, data_orig, mean, var
        xrange = [subid[0],subid[n_elements(subid)-1]]
        plot, subid,data_orig, yrange=[-10*sigma+mean,10*sigma+mean],$
          xrange=xrange, xstyle=1,charsize=1.5,ytitle='Signal'
        xyouts, subid[0], (-7*sigma+mean), 'Channel='+strcompress(ch)
        
        if (n_src ge 1) then begin
            mark=fltarr(n_elements(data))
            mark[*] = -100
            mark[sources]=7*sigma+mean
            oplot, subid, mark,psym=2
        endif

        diff1 = deriv(fader_tsd)+deriv(dipper_tsd)
        diff2 = deriv(diff1)
        sig_diff = sigma * sqrt(2)
        diff = deriv(data)/sig_diff
;        plot, subid, diff1/sigma/sqrt(2),xrange=xrange, xstyle=1,charsize=1.5
 ;       plot, subid, diff2,xrange=xrange, xstyle=1,charsize=1.5
;        plot, subid, diff,xrange=xrange, xstyle=1,charsize=1.5
        
;    plot, noise_tsd, yrange=[-20*sigma,20*sigma] 
;    plot, data, yrange=[-10*sigma+mean,10*sigma+mean] 
;    id = where(flg_big eq 0) 
;    big_tsd = data
;    big_tsd[id]=0
;    plot, big_tsd


        plot, subid,proj, yrange=[-10,10],xrange=xrange, xstyle=1,$
          charsize=1.5,ytitle='Deg. of Scale'
        plot, subid,dipper_tsd,xrange=xrange, xstyle=1,charsize=1.5,$
          ytitle='Signal',yrange=yrange
        plot, subid,fader_tsd,xrange=xrange, xstyle=1,charsize=1.5,$
          ytitle='Signal',yrange=yrange

;plot, data-dipper_tsd-fader_tsd
        wait, 1
;        stop
    endif

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Ploting 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

plot_ps=0

if (plot_ps eq 1) then begin 
;    pson, file='deglitch.eps',/eps
    pson, file='deglitch.ps'
;    isPlot = [1,1,1,1,1] ; only 1 figure is allowed in eps 
    isPlot = [0,0,1,1,0] ; only 1 figure is allowed in eps 
;    isPlot = [0,0,0,1]


;; 0) plot sigma
    if (isPlot[0] eq 1) then begin
        plot,sigma_j/sigma,xtitle='!3MMT scale',ytitle='!4r!3!dj!n/!4r!3'$
                                  ,/ylog, yrange=[0.0001,1]
    endif

;; 1) plot TSD
if (isPlot[1] eq 1) then begin 
    !p.multi=[0,1,3]
    ymax=max(data)
    ymin=min(data)
    yrange=[ymin,ymax]
    yrange=[-sigma*5,sigma*10]
;    yrange=[-0.00005,0.0001]
    plot, data, yrange=yrange
    mark = pspikes
    indx = where(mark eq 0, cnt)
    if (cnt ge 1) then mark[indx] = -100
    oplot, mark*max(data)*1.2, psym=2, symsize=0.5
    mark = nspikes
    indx = where(mark eq 0)
    mark[indx] = -100
    oplot, mark*min(data), psym=2, symsize=0.5
    mark = dippers
    indx = where(mark eq 0)
    mark[indx] = -100
    oplot, mark*min(data), psym=4, symsize=1
    plot, glitchTSD, yrange=yrange
    plot, cleanTSD, yrange=yrange
    !p.multi = 0
endif

;; 2) plot wavelets
objlimit = 10

if (isPlot[2] eq 1) then begin 
    device, ysize=20
    !p.multi=[0,1,n_elements(scale)+1]
;    yrange=[-0.00005,0.0001]

;    for istr=1, nstrs do begin
    for istr=3, 3 do begin

        id = where(struct eq istr, cnt)
        if (cnt ge objlimit) then begin
            yrange=[-sigma*10+mean,sigma*10+mean]
            plot, data_orig, yrange= yrange,ymargin=[2,1], yticks=4

            print, 'istr = ',istr

        ;; Object mark
            mark = mask
            id = where(struct ne istr, cnt)
            if (cnt ne 0) then mark[id] = -1.e9
;       indx = where(mark[*,i] eq 0)
;       mark[indx,i] = -1.e9

            for i=1,n_elements(scale)-1 do begin
;                yrange=$
;              [min(wave[*,i]/sigma_j[i])*1.6,max(wave[*,i]/sigma_j[i])*1.6]
;          plot, wave[*,i]/sigma_j[i],ymargin=[2,1], yticks=4, yrange=yrange

;               yrange= [min(wave[*,i])*1.6,max(wave[*,i])*1.6]
               yrange= [-1.5e-5,7.5e-5]
          plot, wave[*,i],ymargin=[2,1], yticks=4, yrange=yrange
;          oplot, abs(mark[*,i])*max(wave[*,i])*1.6, psym=2, symsize=0.5
          oplot, abs(mark[*,i])*5.e-5, psym=2, symsize=0.5
            endfor
            yrange=[-sigma*10+mean,sigma*10+mean]
            plot, wave[*,0], yrange=yrange,ymargin=[2,1], yticks=4
        endif
    endfor

    !p.multi = 0

endif

;; 3) plot objects
if (isPlot[3] eq 1) then begin 
    device, ysize=12
    !p.multi=[0,2,3]

;    glitch_tsd = fltarr(n_elements(data))
    for istr=1, nstrs do begin
        part_wave=fltarr(n_elements(data),n_elements(scale))
        id = where(struct eq istr, cnt)
        if (cnt ge objlimit) then begin
            print, istr
            part_wave[id] = wave[id]
            mmt_inverse, part_wave, part_tsd, scale=scale            
            plot, part_tsd
            xyouts, 100, (max(part_wave)+min(part_wave))/2, istr
;            glitch_tsd = glitch_tsd + part_tsd
        endif
    endfor

;    !p.multi=[0,1,4]
;    plot, data
;    plot, dipper_tsd
;    plot, fader_tsd
;;    plot, glitch_tsd


    
;    sk_orig = skewness(data)
;    sk_clean = skewness(clean_tsd)
;    print, 'skewness (original) = ',sk_orig 
;    print, 'skewness (cleaned)  = ',sk_clean


;    plot, clean_tsd
;    !p.multi=0
;    plot, proj, ytitle='Number of significant wavelet'    

    !p.multi=0
endif

if (isPlot[4] eq 1) then begin
    icplot, make_squarish(wave),/ax
endif


    psoff
endif 

;stop


endif

jump:

return
end
