; Copyright (c) 2004-2006, ASTRO-F Data Reduction Team. All rights reserved.
;
;+
; NAME:
;   SMPL_MODE
;
; PURPOSE:
;       This function receives packet_ID (and additional informations),
;       then returns a value which indicates sampling sequence pattern.
;
;
; CATEGORY:
;   Function. Green Box library in ASTRO-F pipeline.
;
; CALLING SEQUENCE:
;   Result = smpl_mode(packetid, AFTIME=aftime, TYPE=type)
;
; INPUTS:
;    packetid: Packet ID (scalar or array)
;    aftime  : AFTIME (scalar or array)
;    type    : Data type ('FIS' or 'FISVI')
;
; OUTPUTS:
;
; RETURN:
;   This function returns a SEQ (sampling) pattern indicator.
;      0 : Simple sample mode
;      1 : CDS mode
;      2 : Co-add sample mode
;      3 : High speed sample mode
;      4 : FTS mode
;      5 : High speed sample (quadruple co-add) mode for maneuver operation (Added after the launch.)
;      9 : other modes
;
; MODIFICATION HISTORY:
;       Dec. 2004    Written by Makiuti.
;       Aug. 2006    Last modified
;-
FUNCTION SMPL_MODE, PACKETID, AFTIME=aftime, TYPE=type
;
  ON_ERROR, 2
;
  IF N_PARAMS() EQ 0 THEN MESSAGE, 'Syntax:  result = smpl_mode(packetid, AFTIME=aftime, TYPE=type)'
  IF n_elements(aftime) EQ 0 THEN aftime=600000000; 2019/01/05 10:40:00
  IF n_elements(aftime) GT 1 THEN aftime=aftime[0]
  IF n_elements(type) EQ 0 THEN type='FIS'
;
  ;Make an array for SEQ (sampling) pattern indicator
  seq_indicator = make_array(n_elements(packetid), value=9)

  ; for simulated data
  IF (STRMATCH(type, 'FISVI*')) THEN BEGIN

    ind = where(packetid EQ '40'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '41'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '42'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '43'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode
    ind = where(packetid EQ '50'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '51'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '52'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '53'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode

  ; Old laboratory data : up to 2004/07/26 00:00:00(JST)
  ENDIF ELSE IF (aftime LT 144147620) THEN BEGIN

    ind = where(packetid EQ '40'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '41'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '42'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '43'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode
    ind = where(packetid EQ '50'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 3 ; High speed sample mode
    ind = where(packetid EQ '51'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '52'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '53'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode

  ; Old laboratory data : from 2004/07/26 to 2004/7/30 12:27:00 (JST)
  ENDIF ELSE IF ((aftime GT 144147620) AND (aftime LT 144538020)) THEN BEGIN

    ind = where(packetid EQ '40'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '41'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '42'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '43'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode
    ind = where(packetid EQ '50'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '51'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '52'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '53'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode

  ; Old laboratory data : from 2004/7/30 12:27:00 to 2004/12/2 13:00:00 (JST)
  ENDIF ELSE IF ((aftime GT 144538020) AND (aftime LT 155340000)) THEN BEGIN

    ind = where(packetid EQ '40'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '41'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '42'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '43'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode
    ind = where(packetid EQ '50'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '51'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 3 ; High speed sample mode
    ind = where(packetid EQ '52'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '53'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode

  ; Definitions as at launch time : from 2004/12/2 13:00:00 to 2006/05/20 00:00:00 (JST)
  ENDIF ELSE IF ((aftime GT 155340000) AND (aftime LT 201430800)) THEN BEGIN

    ind = where(packetid EQ '40'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '41'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode
    ind = where(packetid EQ '42'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '43'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 3 ; High speed sample mode
    ind = where(packetid EQ '50'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '51'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode
    ind = where(packetid EQ '52'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '53'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 3 ; High speed sample mode

  ; Latest nominal modes (default) : from 2006/05/20 00:00:00 (JST)
  ENDIF ELSE BEGIN

    ind = where(packetid EQ '40'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '41'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode
    ind = where(packetid EQ '42'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 2 ; Co-add sample mode
    ind = where(packetid EQ '43'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 3 ; High speed sample mode
    ind = where(packetid EQ '50'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 0 ; Simple sample mode
    ind = where(packetid EQ '51'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 1 ; CDS mode
    ind = where(packetid EQ '52'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 5 ; High speed sample mode for maneuver operation (quadruple co-add)
    ind = where(packetid EQ '53'X, cnt)
    IF (cnt NE 0) THEN seq_indicator[ind] = 3 ; High speed sample mode

  ENDELSE

  ; Definition for FTS mode is always same for the time being
  ind = where( (packetid EQ '0E'X) OR (packetid EQ '0F'X) OR (packetid EQ '38'X) OR $
               (packetid EQ '39'X) OR (packetid EQ '3A'X) OR (packetid EQ '3B'X) OR $
               (packetid EQ '3E'X) OR (packetid EQ '3F'X) OR (packetid EQ '4C'X) OR $
               (packetid EQ '4D'X) OR (packetid EQ '78'X) OR (packetid EQ '79'X) OR $
               (packetid EQ '7B'X) OR (packetid EQ '7C'X) OR (packetid EQ '7D'X), cnt)
  IF (cnt NE 0) THEN seq_indicator[ind] = 4 ; FTS mode

  RETURN, seq_indicator

  END
