; $Id: gb_flag_bad_data.pro,v 1.15 2007/01/16 10:14:14 makiuti Exp $
;
; Copyright (c) 2002-2007, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;    GB_Flag_Bad_Data.pro
;
; PURPOSE:
;    Function sets the flags to the data with the following conditions.
;
;    ;;;each pixel condition;;;
;    dead;         1 = dead pixel
;    saturate;     1 = saturated pixel
;    reset;        1 = immediately after reset
;    bad;          1 = bad pixel(for dead, saturate, NaN
;                              blank, in_saa)
;    ;;;each frame condition;;;
;    in_saa;       1 = anonymous sec. after SAA passage
;    near_moon;    1
;    bad_frame;    1 = bad frames (blank, near_moon, in_saa)
;
; CATEGORY:
;   Function. Green Box procedure in ASTRO-F pipeline.
;
; CALLING SEQUENCE:
;    Result = gb_flag_bad_data (tsd
;    [, ccf_deadpix = flag_deadpix, ccf_satpix = flag_satpix
;     , ccf_moon = ccf_moon, ccf_saa = ccf_saa])
;
;    ;; Default values for ccf_* are 1
;
; INPUTS:
;    tsd      TSD object reference.
;    ccf      Calibration constants files.
;
; OUTPUTS:
;    tsd       TSD object reference. Flagged TSD.
;
; RETURN:
;   !SUCCEED if function ended successfully.
;   !FAIL in failure.
;
; RESTRICTIONS:
;
; PROCEDURE:
;    1. Extract detector data, flag and status from TSD.
;    2. Set flag, 'reset' for simple and co-added modes
;    3. Set flags for in_saa (not implemented)
;    4. Set flags for near_moon (not implemented)
;    5. Set bad_frame.
;    6. Set flags for "detector pixel condition|,";
;       dead pixel(dead), and saturation(saturate) defined in CCF.
;    7. Set flags for bad pixels(bad) in "each pixel condition".
;       by examining the other flags.
;    8. Write history in TSD header.
;    9. Output TSD object reference.
;
;
; EXAMPLE:
;
; REFERENCE:
;    Baba et al., "ASTRO-F/FIS Survey Data Structure"
;    Kawamura et al. 2004 Oct. "ASTRO-F Pipeline modules, Green Box"
; MODIFICATION HISTORY:
;    Oct. 2002     Written by A. Kawamura.
;    Dec. 2003     Changed for TSD ver.2
;    Aug. 2004     Updated for TSD ver.3
;    Jan. 2007     Last modified by Makiuti
;-
FUNCTION gb_flag_bad_data, tsd, $
            ccf_deadpix = ccf_deadpix, ccf_satpix = ccf_satpix, $
            ccf_moon = ccf_moon, ccf_saa = ccf_saa

;ON_ERROR 2

;;; Starting GB_Flag_Bad_Data ;;;
    LMESSAGE, 5, 'Executing GB_Flag_Bad_Data...', /CONTINUE

;;; Writing messages ;;;;

    ret = gb_wr_start_msg('gb_flag_bad_data')

;;; Extract values from TSD object ;;;

    vdet = tsd->Get_value_tsd('FIS_OBS','DET')

  ;;; number of channels ;;;
    si = SIZE(vdet)
    num_ch = si[1]
    num_data = si[2]

;;----------------------------------------------------------
;; Check packetid and set flag "reset"
;;----------------------------------------------------------
    packetid = tsd->Get_value_tsd('FIS_OBS','PacketID')
    aftime = tsd->Get_value_tsd('FIS_OBS','AFTIME')
    inst = tsd->Get_header_keyword_tsd('PRIMARY','INSTRUME')

    dummy = packetid - shift(packetid, 1)
    size_dummy = SIZE(dummy)

    dummy = reform(dummy[1:size_dummy[1]-1])
    ind = WHERE(dummy NE 0, cnt)

    ;;-----------------------
    ;; When packetID is constant in a file
    ;;-----------------------
    IF (cnt EQ 0) THEN BEGIN

      msg = "PacketID is constant..."
      LMESSAGE, 7, msg ,/CONTINUE

      ; If the first sample is Simple, Co-add, or Hi-speed mode (not CDS nor FTS)
      IF (smpl_mode(packetid[0], AFTIME=aftime[0], TYPE=inst) EQ 0) OR $
         (smpl_mode(packetid[0], AFTIME=aftime[0], TYPE=inst) EQ 2) OR $
         (smpl_mode(packetid[0], AFTIME=aftime[0], TYPE=inst) EQ 3) OR $
         (smpl_mode(packetid[0], AFTIME=aftime[0], TYPE=inst) EQ 5) THEN BEGIN

        ; Counter status are read.
        msg="Not CDS or FTS mode (Simple, Co-added, or Hi-speed) ..."
        LMESSAGE, 7, msg, /CONTINUE

        IF (num_ch EQ 75) THEN BEGIN ;; LW
            rstctw = tsd->Get_Value_TSD('FIS_OBS','RstCntWideL')
            rstctn = tsd->Get_Value_TSD('FIS_OBS','RstCntN170')
        ENDIF ELSE IF (num_ch EQ 100) THEN BEGIN ;; SW
            rstctw = tsd->Get_Value_TSD('FIS_OBS','RstCntWideS')
            inst=tsd->Get_Header_Keyword_TSD('Primary','Instrume')
            IF (STRMATCH(inst, 'FISVI*')) THEN BEGIN
                LMESSAGE, 7, 'Data is created by Simulator', /CONTINUE
                ;rstctw = rstctw
            ENDIF ELSE BEGIN
                ;rstctw = SHIFT(rstctw,1) ; for past laboratory data
            ENDELSE
            rstctn = tsd->Get_Value_TSD('FIS_OBS','RstCntN60')
        ENDIF ELSE BEGIN
            msg='Channel number is incorrect. Neither LW nor SW?'
            LMESSAGE, 1, msg, /CONTINUE
            RETURN, !FAIL
        ENDELSE

        ; Find count up frames
        rstw = rstctw - SHIFT(rstctw, 1)
        rstn = rstctn - SHIFT(rstctn, 1)

        LMESSAGE, 7, 'Setting flags (reset)...', /CONTINUE

        ; Wide
        ind2 = WHERE(rstw GE 1, cnt2)
        IF (cnt2 GT 1) THEN rstw[ind2] = 1

        ; Narrow
        ind2 = WHERE(rstn GE 1, cnt2)
        IF (cnt2 GT 1) THEN rstn[ind2] = 1

        ;
        IF (num_ch EQ 75) THEN BEGIN ;; LW
          rst = bytarr(num_ch, num_data)
          FOR n = 0, 44 DO BEGIN
            rst[n, *] = rstw
          ENDFOR
          FOR n = 45, 74 DO BEGIN
            rst[n, *] = rstn
          ENDFOR
        ENDIF ELSE BEGIN ;; SW
          rst = bytarr(num_ch, num_data)
          FOR n = 0, 59 DO BEGIN
            rst[n, *] = rstw
          ENDFOR
          FOR n = 60, 99 DO BEGIN
            rst[n, *] = rstn
          ENDFOR
        ENDELSE
        ;; Set reset flag
        err = tsd->Set_Pix_Flag_TSD('FIS_OBS','Reset',rst)

      ; If FTS mode
      ENDIF ELSE IF (smpl_mode(packetid[0], AFTIME=aftime[0], TYPE=inst) EQ 4) THEN BEGIN

        ; Counter status are read.
        msg="FTS mode..."
        LMESSAGE, 7, msg, /CONTINUE

        IF (num_ch EQ 75) THEN BEGIN ;; LW
            rstctw = tsd->Get_Value_TSD('FIS_OBS','RstCntWideL')
            rstctn = tsd->Get_Value_TSD('FIS_OBS','RstCntN170')
        ENDIF ELSE IF (num_ch EQ 100) THEN BEGIN ;; SW
            rstctw = tsd->Get_Value_TSD('FIS_OBS','RstCntWideS')
            rstctn = tsd->Get_Value_TSD('FIS_OBS','RstCntN60')
        ENDIF ELSE BEGIN
            msg='Channel number is incorrect. Neither LW nor SW?'
            LMESSAGE, 1, msg, /CONTINUE
            RETURN, !FAIL
        ENDELSE

        ; Find count up frames
        rstw = rstctw - SHIFT(rstctw, 1)
        rstn = rstctn - SHIFT(rstctn, 1)

        LMESSAGE, 7, 'Setting flags (reset)...', /CONTINUE

        ; Wide
        ind2 = WHERE(rstw GE 1, cnt2)
        IF (cnt2 GT 1) THEN rstw[ind2] = 1

        ; Narrow
        ind2 = WHERE(rstn GE 1, cnt2)
        IF (cnt2 GT 1) THEN rstn[ind2] = 1

        ;
        IF (num_ch EQ 75) THEN BEGIN ;; LW
          rst = bytarr(num_ch, num_data)
          FOR n = 0, 44 DO BEGIN
            rst[n, *] = shift(rstw, 1) ; Shift 1 frame
          ENDFOR
          FOR n = 45, 74 DO BEGIN
            rst[n, *] = rstn
          ENDFOR
        ENDIF ELSE BEGIN ;; SW
          rst = bytarr(num_ch, num_data)
          FOR n = 0, 59 DO BEGIN
            rst[n, *] = shift(rstw, 2) ; Shift 2 frames
          ENDFOR
          FOR n = 60, 99 DO BEGIN
            rst[n, *] = rstn
          ENDFOR
        ENDELSE
        ;; Set reset flag
        err = tsd->Set_Pix_Flag_TSD('FIS_OBS','Reset',rst)


      ENDIF ELSE IF (smpl_mode(packetid[0], AFTIME=aftime[0], TYPE=inst) EQ 9) THEN BEGIN
        msg = "Unknown packeID only..."
        LMESSAGE, 1, msg, /CONTINUE
        RETURN, -1
      ENDIF

    ;;-----------------------
    ;; When packetID changes in a file
    ;;-----------------------
    ENDIF ELSE BEGIN

      msg='Different PACKETID in TSD...'
      LMESSAGE, 7, msg, /CONTINUE
      rst=bytarr(num_ch,num_data)

      ; Counter status are read.
      IF num_ch EQ 75 THEN BEGIN;; LW
        rstctw = tsd->Get_Value_TSD('FIS_OBS','RstCntWideL')
        rstctn = tsd->Get_Value_TSD('FIS_OBS','RstCntN170')
      ENDIF ELSE IF num_ch EQ 100 THEN BEGIN;; SW
        rstctw = tsd->Get_Value_TSD('FIS_OBS','RstCntWideS')
        inst=tsd->Get_Header_Keyword_TSD('Primary','Instrume')

        IF (STRMATCH(inst, 'FISVI*')) THEN BEGIN
          LMESSAGE, 7, 'Data is created by Simulator', /CONTINUE
          ;rstctw = rstctw
        ENDIF ELSE BEGIN
          ;rstctw = SHIFT(rstctw,1) ; for past laboratory data
        ENDELSE
        rstctn = tsd->Get_Value_TSD('FIS_OBS','RstCntN60')
      ENDIF ELSE BEGIN
        msg='Channel number is incorrect. Neither LW nor SW?'
        LMESSAGE, 1, msg, /CONTINUE
        RETURN, !FAIL
      ENDELSE

      ; Find count up frames
      rstw = rstctw - SHIFT(rstctw, 1)
      rstn = rstctn - SHIFT(rstctn, 1)

      LMESSAGE, 7, 'Setting flags (reset)...', /CONTINUE

      ; Wide
      ind2 = WHERE(rstw GE 1, cnt2)
      IF (cnt2 GT 1) THEN rstw[ind2] = 1

      ; Narrow
      ind2 = WHERE(rstn GE 1, cnt2)
      IF (cnt2 GT 1) THEN rstn[ind2] = 1


      FOR i = 0, cnt-1 DO BEGIN

        ; If mode before a mode-change is Simple, Co-add, or Hi-speed mode (not CDS nor FTS)
        IF (smpl_mode(packetid[ind[i]], AFTIME=aftime[ind[i]], TYPE=inst) EQ 0) OR $
           (smpl_mode(packetid[ind[i]], AFTIME=aftime[ind[i]], TYPE=inst) EQ 2) OR $
           (smpl_mode(packetid[ind[i]], AFTIME=aftime[ind[i]], TYPE=inst) EQ 3) OR $
           (smpl_mode(packetid[ind[i]], AFTIME=aftime[ind[i]], TYPE=inst) EQ 5) THEN BEGIN

          IF i NE 0 THEN BEGIN
            startp = ind[i-1]+1
          ENDIF ELSE BEGIN
            startp = 0
          ENDELSE
            endp = ind[i]

          IF (num_ch EQ 75) THEN BEGIN
            FOR n = 0, 44 DO BEGIN
              rst[n, startp:endp] = rstw[startp:endp]
            ENDFOR
            FOR n = 45, 74 DO BEGIN
              rst[n, startp:endp] = rstn[startp:endp]
            ENDFOR
          ENDIF ELSE BEGIN
            FOR n = 0, 59 DO BEGIN
              rst[n, startp:endp] = rstw[startp:endp]
            ENDFOR
            FOR n = 60, 99 DO BEGIN
              rst[n, startp:endp] = rstn[startp:endp]
            ENDFOR
          ENDELSE

        ; If FTS mode
        ENDIF ELSE IF (smpl_mode(packetid[ind[i]], AFTIME=aftime[ind[i]], TYPE=inst) EQ 4) THEN BEGIN

          IF i NE 0 THEN BEGIN
            startp = ind[i-1]+1
          ENDIF ELSE BEGIN
            startp = 0
          ENDELSE
            endp = ind[i]

          IF (num_ch EQ 75) THEN BEGIN
            FOR n = 0, 44 DO BEGIN
              rst[n, startp:endp] = shift(rstw[startp:endp], 1) ; Shift 1 frame
            ENDFOR
            FOR n = 45, 74 DO BEGIN
              rst[n, startp:endp] = rstn[startp:endp]
            ENDFOR
          ENDIF ELSE BEGIN
            FOR n = 0, 59 DO BEGIN
              rst[n, startp:endp] = shift(rstw[startp:endp], 2) ; Shift 2 frames
            ENDFOR
            FOR n = 60, 99 DO BEGIN
              rst[n, startp:endp] = rstn[startp:endp]
            ENDFOR
          ENDELSE

        ; If unknown mode
        ENDIF ELSE IF (smpl_mode(packetid[ind[i]], AFTIME=aftime[ind[i]], TYPE=inst) EQ 9) THEN BEGIN
          PRINT,packetid[ind[i]]
          msg = "Unknown packeID found..."
          LMESSAGE, 5, msg, /CONTINUE
          ;RETURN, -1
        ENDIF

      ENDFOR

      ; If last mode in the file is Simple or Co-add or Hi-speed mode (not CDS)
      IF (smpl_mode(packetid[ind[cnt-1]+1], AFTIME=aftime[ind[cnt-1]+1], TYPE=inst) EQ 0) OR $
         (smpl_mode(packetid[ind[cnt-1]+1], AFTIME=aftime[ind[cnt-1]+1], TYPE=inst) EQ 2) OR $
         (smpl_mode(packetid[ind[cnt-1]+1], AFTIME=aftime[ind[cnt-1]+1], TYPE=inst) EQ 3) OR $
         (smpl_mode(packetid[ind[cnt-1]+1], AFTIME=aftime[ind[cnt-1]+1], TYPE=inst) EQ 5) THEN BEGIN

        startp = ind[cnt-1]+1
        endp = num_data-1

        IF num_ch EQ 75 THEN BEGIN
          FOR n = 0, 44 DO BEGIN
            rst[n, startp:endp] = rstw[startp:endp]
          ENDFOR
          FOR n = 45, 74 DO BEGIN
            rst[n, startp:endp] = rstn[startp:endp]
          ENDFOR
        ENDIF ELSE BEGIN
          FOR n = 0, 59 DO BEGIN
            rst[n, startp:endp] = rstw[startp:endp]
          ENDFOR
          FOR n = 60, 99 DO BEGIN
            rst[n, startp:endp] = rstn[startp:endp]
          ENDFOR
        ENDELSE

      ; If FTS mode
      ENDIF ELSE IF (smpl_mode(packetid[ind[cnt-1]+1], AFTIME=aftime[ind[cnt-1]+1], TYPE=inst) EQ 4) THEN BEGIN

        startp = ind[cnt-1]+1
        endp = num_data-1

        IF num_ch EQ 75 THEN BEGIN
          FOR n = 0, 44 DO BEGIN
            rst[n, startp:endp] = shift(rstw[startp:endp], 1) ; Shift 1 frame
          ENDFOR
          FOR n = 45, 74 DO BEGIN
            rst[n, startp:endp] = rstn[startp:endp]
          ENDFOR
        ENDIF ELSE BEGIN
          FOR n = 0, 59 DO BEGIN
            rst[n, startp:endp] = shift(rstw[startp:endp], 2) ; Shift 2 frames
          ENDFOR
          FOR n = 60, 99 DO BEGIN
            rst[n, startp:endp] = rstn[startp:endp]
          ENDFOR
        ENDELSE

      ; If unknown mode
      ENDIF ELSE IF (smpl_mode(packetid[ind[cnt-1]+1], AFTIME=aftime[ind[cnt-1]+1], TYPE=inst) EQ 9) THEN BEGIN
         print,packetid[ind[cnt-1]+1]
         msg = "Unknown packeID found..."
         LMESSAGE, 5, msg, /CONTINUE
        ;RETURN, -1
      ENDIF
      ;; Set reset flag
      err = tsd->Set_Pix_Flag_TSD('FIS_OBS', 'Reset', rst)

    ENDELSE

;;----------------------------------------------------------
;; Set Frame Flags
;;----------------------------------------------------------
LMESSAGE, 7, 'Setting frame flags...', /CONTINUE

  ;;-----------------------
  ;; Read statuses about SAA, Polar, Moon, and BiasBoost
  ;;-----------------------
  se_in_saa = tsd->Get_Status_TSD('SE','In_SAA')
  se_in_polar = tsd->Get_Status_TSD('SE','In_Polar')
  tm_saa = tsd->Get_Value_TSD('SE','Tm_SAA')
  pr_aa_lun = tsd->Get_Value_TSD('PR','AA_Lun')

  IF num_ch EQ 75 THEN BEGIN
    LMESSAGE, 7, "Reading LW Bias Boost status...", /CONTINUE
    biasboost = tsd->Get_Status_TSD('FIS_OBS','LwBoostOn')
  ENDIF ELSE BEGIN
    LMESSAGE, 7, "Reading SW Bias Boost status...", /CONTINUE
    biasboost = tsd->Get_Status_TSD('FIS_OBS','SwBoostOn')
  ENDELSE

  ;;-----------------------
  ;; Set frame flags 'in_saa' and 'near_moon'
  ;;-----------------------
  insaa = tsd->Get_Flag_TSD('FIS_OBS','In_SAA')
  nmoon = tsd->Get_Flag_TSD('FIS_OBS','Near_Moon')

  ; Here, reflect statuses gotten from 'SE' and 'PR' extensions

  ; Set SAA flag at frames where tm_saa has negative value
  idx = where(tm_saa LT 0.0, cnt_saa)
  IF (cnt_saa GT 0) THEN insaa[idx] = 1B

  err = tsd->Set_Flag_TSD('FIS_OBS','In_SAA', insaa)
  err = tsd->Set_Flag_TSD('FIS_OBS','Near_Moon', nmoon)

  ;;-----------------------
  ;; Set frame flags 'undef_anom_frame' and 'bad_frame'
  ;;-----------------------
  ;;; reading frame flags
      blank = tsd->Get_Flag_TSD('FIS_OBS','Blank')
      undef_anom_frame = tsd->Get_Flag_TSD('FIS_OBS','Undef_Anom_Frame')

  ;;; setting undef_anom_frame flag due to bias boost
      LMESSAGE, 7, "Setting undef_anom_frame flag due to Bias Boost...", /CONTINUE
      ind_biasboost=WHERE(biasboost NE 0, cnt_biasboost)

      IF cnt_biasboost NE 0 THEN undef_anom_frame[ind_biasboost]=1
      ret = tsd->Set_Flag_TSD('FIS_OBS','Undef_Anom_Frame', undef_anom_frame)

  ;;; making bad flag arrays
      bad_frame = tsd->Get_Flag_TSD('FIS_OBS','Bad_Frame')
      dummy = insaa + nmoon + blank + undef_anom_frame
      ind = WHERE (dummy NE 0, cnt)
      IF cnt NE 0 THEN dummy[ind]=1

      msg = "Setting flag, BAD_FRAME..."
      LMESSAGE, 7, msg, /CONTINUE

      err = tsd->Set_Flag_TSD('FIS_OBS','Bad_Frame', dummy)

;;----------------------------------------------------------
;; Set pixel flags for "detector pixel condition"
;;----------------------------------------------------------
;
;      dead pixel (dead), and saturation (saturate)

  ;;-----------------------
  ;; Set flags for dead pixels (dead) in "each pixel condition".
  ;; (Dead pixel is defined in CCF)
  ;;-----------------------
  IF (N_elements(ccf_deadpix) EQ 0) THEN ccf_deadpix = 1 ;#default value

  IF (ccf_deadpix NE 0) THEN BEGIN

    LMESSAGE, 7, 'Setting flags (dead)...', /CONTINUE
     ;;; Read CCF ;;;
     ;;;    CCF data is an array with the number of pixels.
     ;;;    1 for dead pixels, 0 for healthy pixels.
    LMESSAGE, 7, "Read CCF for dead pixel", /CONTINUE
    IF (num_ch EQ 100) THEN BEGIN ;; SW
      deadpix = CCF_Read_CCF('DEADS', tsd, HEADER=hdr, ERROR=err)
    ENDIF ELSE IF (num_ch EQ 75) THEN BEGIN ;; LW
      deadpix = CCF_Read_CCF('DEADL', tsd, HEADER=hdr, ERROR=err)
    ENDIF

  ENDIF ELSE BEGIN

    LMESSAGE, 7, "CCF for dead pixels was not read.", /CONTINUE
    deadpix = bytarr(num_ch)

  ENDELSE

  ;; Display just for information
  IF (num_ch EQ 100) THEN BEGIN
    LMESSAGE, 7, "Dead pix. information for SW...", /CONTINUE
    LMESSAGE, 7, reform(deadpix, 20,5), /PRINT
  ENDIF ELSE IF (num_ch EQ 75) THEN BEGIN
    LMESSAGE, 7, "Dead pix. information for LW...", /CONTINUE
    LMESSAGE, 7, reform(deadpix, 15,5), /PRINT
  ENDIF

  fdead = tsd->Get_pix_flag_tsd('FIS_OBS', 'Dead')

  ind = WHERE(deadpix NE 0, cnt)
  IF cnt NE 0 THEN BEGIN
    FOR i = 0, cnt-1 DO BEGIN
      fdead[ind[i], *] = 1
    ENDFOR
  ENDIF

  err = tsd->Set_pix_flag_tsd('FIS_OBS','Dead', fdead)


  ;;-----------------------
  ;; Set flags for saturated pixels (saturate) in "each pixel condition".
  ;  (Parameters is given in CCF)
  ;;-----------------------
  IF (N_elements(ccf_satpix) EQ 0) THEN ccf_satpix = 1 ;#default value

  IF (ccf_satpix NE 0) THEN BEGIN

    LMESSAGE, 7, 'Setting flags (saturate)...', /CONTINUE
     ;;; Read CCF ;;;
     ;;;    CCF data is a constant.
     ;;;    The unit is Volt.
    LMESSAGE, 7, "Read CCF for saturation pixel", /CONTINUE
    IF (num_ch EQ 100) THEN BEGIN ;; SW
      satu = CCF_Read_CCF('SATUS', tsd, HEADER=hdr, ERROR=err)
    ENDIF ELSE IF (num_ch EQ 75) THEN BEGIN ;; LW
      satu = CCF_Read_CCF('SATUL', tsd, HEADER=hdr, ERROR=err)
    ENDIF

  ENDIF ELSE BEGIN

    LMESSAGE, 7, "CCF for saturated pixels was not read.", /CONTINUE
    satu = 1000000.D ; virtually no limit

  ENDELSE

  ;; Display just for information
  msg = "Satulation level : " + string(satu)
  LMESSAGE, 7, msg, /PRINT

  fsat = tsd->Get_pix_flag_tsd('FIS_OBS','Saturate')

  dummy_adu = (satu+9.34e-5)/(1.416e-6)

  ;;; Check co-add mode ;;;
    ; Get SEQ (sampling) pattern indicator array
    seq_mode = smpl_mode(packetid, AFTIME=aftime[0], TYPE=inst)
    ; If Co-add mode
    ind = WHERE(seq_mode EQ 2, cnt)
    IF cnt NE 0 THEN BEGIN
      vdet[*,ind] = vdet[*,ind]/6.D
    ENDIF

  ind=WHERE(vdet GT dummy_adu, cnt)
  IF cnt GE 1 THEN fsat[ind] = 1

  err = tsd->Set_pix_flag_tsd('FIS_OBS','Saturate', fsat)

  ;;-----------------------
  ;; Set flags for "bad"
  ;;-----------------------
  fbad = tsd->Get_pix_flag_tsd('FIS_OBS','Bad')

  ;;; reading pixel flags
      dead = tsd->Get_pix_flag_tsd('FIS_OBS','Dead')
      sat = tsd->Get_pix_flag_tsd('FIS_OBS','Saturate')

  ;;; reading frame flags
      bad_frame = tsd->Get_flag_tsd('FIS_OBS','Bad_Frame')
      fbad_frame = fbad
      fbad_frame[*,*] = 0.

      ind = WHERE(bad_frame NE 0, cnt)
      IF cnt NE 0 THEN BEGIN
        FOR i = 0L, cnt-1 DO BEGIN
          fbad_frame[*,ind[i]] =1
        ENDFOR
      ENDIF

  ;;; making flag arrays
      dummy = dead+sat+fbad_frame
      ind = WHERE (dummy NE 0, cnt)
      IF cnt NE 0 THEN dummy[ind]=1

      err = tsd->Set_pix_flag_tsd('FIS_OBS','Bad', dummy)


;;----------------------------------------------------------
;; Write a history in TSD header
;;----------------------------------------------------------
   LMESSAGE, 7, 'Writing HISTORY', /CONTINUE
   name_ver=gb_get_name_ver(infile='gb_flag_bad_data')
   hist_message = STRCOMPRESS(' '+name_ver+' done. CCF= ')

   ;IF KEYWORD_SET(ccf) NE 1 THEN BEGIN
   ;   hist_message = hist_message + 'NONE'
   ;ENDIF ELSE BEGIN
   ;   hist_message = hist_message+ccf
   ;ENDELSE

   dummy = tsd_fxaddhist(tsd, hist_message, /prres)
   IF dummy NE 0 THEN $
     LMESSAGE, 3, 'Unable to write HISTORY', /CONTINUE

   return, !SUCCEED

END
