; $Id: gb_flux_calib.pro,v 1.9 2006/01/18 06:57:05 baba Exp $
;
; Copyright (c) 2002-2005, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;   gb_flux_calib.pro
; PURPOSE:
;   The flux is calibrated based on the calibration parameters
;   provided by the calibration team.
; CATEGORY:
;   Function. Green Box procedure in ASTRO-F pipeline.
;
; CALLING SEQUENCE:
;   Result = gb_flux_calib(tsd
;   [, ccf_calib = flux_calib_sf.ccf, ccf_badpix = flux_calib_badpix.ccf])
;
; INPUTS:
;   tsd    TSD object reference.
;     Keywords:
;       ccf_calib    Calibration constants file for flux calibration;
;       ccf_badpix   Calibration constants file for bad pixel for calibration;
;
; OUTPUTS:
;   tsd    TSD object reference. Flux-calibrated TSD.
;
; RETURN:
;   !SUCCEED if function ended successfully.
;   !FAIL in failure.
;
; RESTRICTIONS:
;
; PROCEDURE:
;   1. Detector data is read.
;   2. Calibration constants files are read.
;   3. Conversion factor is determined; the conversion factor
;   should depend on the detector output.
;   4. Calibration is carried out.
;   5. Calibrated TSD with history in its header is written.
;
;   Calibration parameters are to be determined by the calibration team.
;
; EXAMPLE:
;
; REFERENCE:
;   Baba et al. 2002 Nov. 21, ``ASTRO-F/FIS Survey Data Structure"
;   Kawamura et al. 2004 Mar. ``ASTRO-F/FIS Pipeline - Green Box - "
;
; MODIFICATION HISTORY:
;   Oct. 2002    Written by A. Kawamura.
;   Nov. 2005    Last modified by Makiuti
;-

FUNCTION gb_flux_calib, tsd, ccf_calib = ccf_calib, ccf_badpix = ccf_badpix

;ON_ERROR 2

;;; Starting GB_Flux_Calib ;;;
    LMESSAGE, 5, 'Executing gb_flux_calib...', /CONTINUE

;;; Extract values from TSD object reference ;;;
    in_flux = tsd->Get_value_tsd('FIS_OBS','Flux')
    ret = af_finite(in_flux, flux)
    ferr = tsd->Get_value_tsd('FIS_OBS','FErr')

    ;;; number of channels ;;;
    si = SIZE(flux)
    num_ch = si[1]
    num_data = si[2]

    ;;-----------------------
    ;; Read CCFs (calibration constants files)
    ;;-----------------------

    ;;;;;
    ;; Derive conversion (scaling) factors
    ;;;;;
    IF (N_elements(ccf_calib) EQ 0) THEN ccf_calib = 1 ;#default value

    ;; Get scaling factors
    IF (ccf_calib NE 0) THEN BEGIN ;; Read CCF

      ;;; read CCF ;;;
      LMESSAGE, 7, "Read CCF for scaling factor from Volt to Flux(Jy).", /CONTINUE
      v2fl_ccf = CCF_Read_CCF('V2FL_', tsd, HEADER=hdr, ERROR=err)

      inst=tsd->Get_header_keyword_tsd('Primary','instrume')
      IF (STRMATCH(inst, 'FISVI*')) THEN BEGIN ; for simulated data
        ; get scaling factors
        sf_n60   = v2fl_ccf.sf_sim[0]
        sf_wides = v2fl_ccf.sf_sim[1]
        sf_widel = v2fl_ccf.sf_sim[2]
        sf_n170  = v2fl_ccf.sf_sim[3]
      ENDIF ELSE BEGIN                         ; for real (Lab) data
        sf_n60   = v2fl_ccf.sf_lab[0]
        sf_wides = v2fl_ccf.sf_lab[1]
        sf_widel = v2fl_ccf.sf_lab[2]
        sf_n170  = v2fl_ccf.sf_lab[3]
      ENDELSE

    ENDIF ELSE BEGIN ;; Dose not read CCF, use dummy scaling factors
        sf_n60   = 1.0
        sf_wides = 1.0
        sf_widel = 1.0
        sf_n170  = 1.0
    ENDELSE

    ;; Display just for information
    msg = ["Multiplying scaling factors from Volt to Flux(Jy) : ", string([sf_n60, sf_wides, sf_widel, sf_n170])]
    LMESSAGE, 7, msg, /PRINT

    ;;;;;
    ;; Derive positions of bad pixels
    ;;;;;
    IF (N_elements(ccf_badpix) EQ 0) THEN ccf_badpix = 1 ;#default value

    ;; Get badpix info.
    IF (ccf_badpix NE 0) THEN BEGIN ;; Read CCF

      ;;; read CCF ;;;
      ;;;    CCF data is an array with the number of pixels.
      ;;;    1 for bad pixels, 0 for healthy pixels.
      LMESSAGE, 7, "Read CCF of bad pixels.", /CONTINUE
      IF (num_ch EQ 100) THEN BEGIN ;; SW
        badpix = CCF_Read_CCF('BADPS', tsd, HEADER=hdr, ERROR=err)
      ENDIF ELSE IF (num_ch EQ 75) THEN BEGIN ;; LW
        badpix = CCF_Read_CCF('BADPL', tsd, HEADER=hdr, ERROR=err)
      ENDIF

    ENDIF ELSE BEGIN ;; Dose not read CCF, use dummy scaling factors

      LMESSAGE, 7, "CCF for bad pixels was not read.", /CONTINUE
      badpix = intarr(num_ch)

    ENDELSE

    ;; Display just for information
    msg = ["Bad pixels : ", string(badpix)]
    LMESSAGE, 7, msg, /PRINT

;;----------------------------------------------------------
;; Calibration
;;----------------------------------------------------------
    cal_flux = flux
    cal_ferr = ferr

    IF (num_ch EQ 75) THEN BEGIN
      cal_flux[0:44,*]  = flux[0:44,*] * sf_widel
      cal_flux[45:74,*] = flux[45:74,*] * sf_n170
      cal_ferr[0:44,*]  = ferr[0:44,*] * sf_widel
      cal_ferr[45:74,*] = ferr[45:74,*] * sf_n170
    ENDIF ELSE BEGIN
      cal_flux[0:59,*]  = flux[0:59,*] * sf_wides
      cal_flux[60:99,*] = flux[60:99,*] * sf_n60
      cal_ferr[0:59,*]  = ferr[0:59,*] * sf_wides
      cal_ferr[60:99,*] = ferr[60:99,*] * sf_n60
    ENDELSE

;;----------------------------------------------------------
;; Set flags and quality indicators
;;----------------------------------------------------------

    ;;-----------------------
    ;; Get quality indicators
    ;;-----------------------
    qual_fx_corr  = tsd->Get_quality_tsd('FIS_OBS','Qual_fx_corr')
    qual_fx_param = tsd->Get_quality_tsd('FIS_OBS','Qual_fx_param')

    ;;-----------------------
    ;; Set flag of "no_abscal" by CCF info of 'bad pixel'
    ;;-----------------------
    msg="Setting flag, no_abscal..."
    LMESSAGE, 7, msg, /CONTINUE

    no_abscal = tsd->Get_pix_flag_tsd('FIS_OBS','No_AbsCal')

    badpix_arr = no_abscal
    badpix_arr[*,*] = 0.
    ind_badpix = WHERE(badpix NE 0, cnt)
    IF (cnt NE 0) THEN badpix_arr[ind_badpix,*] = 1

    temp_flag = no_abscal + badpix_arr
    ind = WHERE(temp_flag NE 0, cnt)
    IF (cnt NE 0) THEN temp_flag[ind] = 1

    ret = tsd->Set_pix_flag_tsd('FIS_OBS','No_AbsCal', temp_flag)

    ;;-----------------------
    ;; Set flag of "arith_err"
    ;;-----------------------
    msg = "Setting flag, arith_err"
    LMESSAGE, 7, msg, /CONTINUE

    ind_nan = WHERE(FINITE(cal_flux, /NAN), count_nan)
    ind_inf = WHERE(FINITE(cal_flux, /INFINITY), count_inf)

    msg = STRCOMPRESS('Number of NaN ... '+STRING(count_nan))
    LMESSAGE, 7, msg, /CONTINUE
    msg = STRCOMPRESS('Number of INFINITE ... '+STRING(count_inf))
    LMESSAGE, 7, msg, /CONTINUE

    arith_err = bytarr(num_ch,num_data)
    IF (count_nan NE 0) THEN arith_err[ind_nan] = 1
    IF (count_inf NE 0) THEN arith_err[ind_inf] = 1
    ;ind = WHERE(arith_err NE 0, cnt)
    ;IF (cnt NE 0) THEN arith_err[ind] = 1

    ret = tsd->Set_pix_flag_tsd('FIS_OBS','Arith_Err', arith_err)

    ;;-----------------------
    ;; Set flag of "bad"
    ;;-----------------------
    msg = "Setting flag, bad..."
    LMESSAGE, 7, msg, /CONTINUE

    bad = tsd->Get_pix_flag_tsd('FIS_OBS','Bad')
    temp_flag = no_abscal + arith_err + bad
    ind = WHERE(temp_flag NE 0, cnt)
    IF (cnt NE 0) THEN BEGIN
      temp_flag[ind] = 1
      qual_fx_corr[ind]  = 3  ; set quality
      qual_fx_param[ind] = 3
    ENDIF

    ret = tsd->Set_pix_flag_tsd('FIS_OBS','Bad', temp_flag)

    ;;-----------------------
    ;; Set quality indicators
    ;;-----------------------
    ret = tsd->Set_quality_tsd('FIS_OBS','Qual_fx_corr', qual_fx_corr)
    ret = tsd->Set_quality_tsd('FIS_OBS','Qual_fx_param', qual_fx_param)

;;----------------------------------------------------------
;; Write 'flux' into TSD
;;----------------------------------------------------------
    ret = tsd->Set_value_tsd('FIS_OBS','Flux', cal_flux)
    ret = tsd->Set_value_tsd('FIS_OBS','FErr', cal_ferr)
    LMESSAGE, 7, 'Writing flux calibrated TSD...', /CONTINUE

;;----------------------------------------------------------
;; Write HISTORY in header
;;----------------------------------------------------------
   LMESSAGE, 7, 'Writing HISTORY ', /CONTINUE
   name_ver = gb_get_name_ver(infile='gb_flux_calib')
   hist_message = STRCOMPRESS(' '+name_ver+' done. CCF= ')

   IF KEYWORD_SET(ccf) NE 1 THEN BEGIN
      hist_message = hist_message + 'NONE'
   ENDIF ELSE BEGIN
      hist_message = hist_message+ccf
   ENDELSE

   dummy = tsd_fxaddhist(tsd, hist_message,/prres)
   IF (dummy NE 0) THEN $
     LMESSAGE, 3, 'Unable to write HISTORY', /CONTINUE

   RETURN, !SUCCEED

END
