; $Id: gb_set_glitch_status_gp.pro,v 1.6 2006/05/04 02:37:45 do Exp $
;
; Copyright (c) 2002, ASTRO-F Data Reduction Team.  All rights reserved.
;
;+
; NAME:
;	GB_Set_Glitch_Status.pro
; PURPOSE:
;	Detect spikes by cosmic-ray hits. Flag, "verb|spike," is set.
; CATEGORY:
;	Function. Green Box procedure in ASTRO-F pipeline.
; CALLING SEQUENCE:
;	Result = GB_Set_Glitch_Status(tsd
;	[, CAL_CONST_FILE = cal_const_file, method=method
;	    glitchlist=glitchlist, pixel=pixel, plot=plot, $
;	    _extra=_extra] )
; INPUTS:
;	tsd		TSD object reference.
;	cal_const_file	Calibration constants file;
;			format, usage are TBD (optional.)
;	method		choice between methods
;			1 --> gb_gauss_glitch
;			2 --> gb_bayes_glitch  (default)
;	pixel		operate on these pixels only
;			default: all
;	plot		make plots
;	_extra		for plotting.
; OUTPUTS:
;	tsd		TSD object reference. Deglitched TSD.
;	glitchlist	list of detected glitches.
;	How to manage the summaries is under considerations.
;RETURN
;	!SUCCEED if function ended successfully.
;	!FAIL in failure.
; RESTRICTIONS:
;
; PROCEDURE:
;	1. This function will be provided by the European team.
;	2. The after effects are corrected with the transient correction.
;	3. A large spike may cause saturation.
;	4. It is to be noted that the flux data necessary for
;	the transient
;	correction is lost until the next reset when a detector is
;	saturated.
;	5. The data of previous missions, such as IRTS/FILM and ISO/LWS,
;	6. PHOT, are useful example to investigate the real cosmic-ray
;	glitches on the same type of detectors.
; EXAMPLE:
;
; REFERENCE:
;	Baba et al. 2002 Nov. 21, ``ASTRO-F/FIS Survey Data Structure"
;
; MODIFICATION HISTORY:
;	Oct. 2002		Written by A. Kawamura.
;	Dec. 2002		Header is added.
;	2003-2005   DK		developments
;	June 2005   DK		make default method = 2
;-

FUNCTION GB_Set_Glitch_Status_GP, tsd, Calib_Const_File = ccf, $
	    glitchlist=glitchlist, method=method, pixel=pixel, plot=plot, $
	    _extra=_extra, resetanom=resetanom

;   ON_ERROR 2

    syst = systime(1)

;;; Starting GB_Set_Glitch_Status ;;;
    lmessage, 7,'Executing GB_Set_Glitch_Status...'

    if ( n_elements( method ) EQ 0 ) then method = 2

;;; Extract values from TSD object reference ;;;
    flux = tsd->get_value_tsd( 'fis_obs', 'flux' )
    infl = tsd->get_pix_flag_tsd( 'fis_obs', 'dead' ) OR $
	   tsd->get_pix_flag_tsd( 'fis_obs', 'saturate' ) OR $
	   tsd->get_pix_flag_tsd( 'fis_obs', 'reset' ) OR $
	   tsd->get_pix_flag_tsd( 'fis_obs', 'no_diff' )

; Remove calibration flashes as they are flagged sometimes as glitches.
; TBC later with more realistic data (non-square calibration flashes)
    type = tsd->get_datatype()
    calstat = 0
    if ( type EQ 'FIS_LW' ) then $
	calstat = tsd->get_status_tsd( 'fis_obs', 'calalon' )
    if ( type EQ 'FIS_SW' ) then $
	calstat = tsd->get_status_tsd( 'fis_obs', 'calason' )

;;; number of channels ;;;
    si = SIZE(flux)
    arr_dim = si[0]
    ndet = si[1]
    num_data = si[2]

    flag = bytarr( ndet, num_data )		; for storing the gp-flags
    cntr = intarr( ndet, num_data ) - 1 	; no glitches yet.
    qual = bytarr( ndet, num_data )

    if ( n_elements( pixel ) EQ 0 ) THEN pixel = indgen( ndet ) + 1 $
    else ndet = n_elements( pixel )

    if ( n_elements( glitchlist ) GT 0 ) then begin
	if ( size( glitchlist, /type ) EQ 7 ) then listfile = glitchlist $
	else begin
	    infile = tsd->get_filename()
	    str_len = strpos( infile, ".fits" )
	    listfile = strmid( infile, 0, str_len ) + '_glitch_list.dat'
	endelse
	openw, glunit, listfile, /get_lun
    endif

    pltype = size( plot, /type )
    if ( pltype GT 0 ) then begin
	do_plot=1
	if ( pltype EQ 7 ) then start_hc, plot
    endif

    if ( n_elements( ccf ) EQ 0 ) then ccf = 'GLGP_'
    glitcal = ccf_read_ccf( ccf, tsd, error=err )
    if ( err EQ !FAIL ) then $
	message, "No CCF file found for GP glitch detection", /continue

; read instrument status and covert to one int
; TBC what statuses actually change the input flux
    stat = fix( tsd->get_status_tsd( 'fis_obs', 'shtop' ) ) OR $
	   ishft( tsd->get_status_tsd( 'fis_obs', 'fwposon' ), 1 ) OR $
	   ishft( tsd->get_status_tsd( 'fis_obs', 'mposon' ), 2 ) OR $
	   ishft( tsd->get_status_tsd( 'fis_obs', 'creon' ), 3 )

    if ( type EQ 'FIS_LW' ) then begin
	stat = stat OR $
	       ishft( tsd->get_status_tsd( 'fis_obs', 'calalon' ), 4 ) OR $
	       ishft( tsd->get_status_tsd( 'fis_obs', 'lwbooston' ), 5 ) OR $
	       ishft( tsd->get_status_tsd( 'fis_obs', 'lwbiason' ), 6 ) OR $
	       ishft( tsd->get_status_tsd( 'fis_obs', 'sinalon' ), 7 )
    endif
    if ( type EQ 'FIS_SW' ) then begin
	stat = stat OR $
	       ishft( tsd->get_status_tsd( 'fis_obs', 'calason' ), 4 ) OR $
	       ishft( tsd->get_status_tsd( 'fis_obs', 'swbooston' ), 5 ) OR $
	       ishft( tsd->get_status_tsd( 'fis_obs', 'swbiason' ), 6 ) OR $
	       ishft( tsd->get_status_tsd( 'fis_obs', 'sinason' ), 7 )
    endif
    stat = stat OR $
	   ishft( fix( tsd->get_status_tsd( 'fis_obs', 'calbon' ) ), 8 )
;    ustat = stat[ uniq( stat, sort( stat ) ) ]
;    print, size( stat )
;    print, n_elements( ustat ), ustat, format='(10i10)'

    FOR I = 0, ndet-1 DO BEGIN
      det = pixel[i]
      kd = det - 1

      if ( method EQ 1 ) then begin
	    inflag = infl[kd,*] OR calstat
	    flag[kd,*] = gb_gauss_glitch( flux[kd,*], inflag, glitcal, $
		    pixel=det, seed=seed, $
		    plot=do_plot, unit=glunit, _extra=_extra )
      endif else begin
	    detq = bytarr( num_data )
	    flag[kd,*] = gb_bayes_glitch( flux[kd,*], infl[kd,*], stat, $
		    glitcal, qual=detq, pixel=det, seed=seed, $
		    plot=do_plot, unit=glunit, _extra=_extra )
	    qual[kd,*] = detq
      endelse

      bf = where( flag[kd,*] GT 0 AND flag[kd,*] LT 5, nbf )
      bf = [bf, num_data]
      maxint = 32768
      for kf = 0, nbf - 1 do begin
	cc = bf[kf+1] - bf[kf]
	cntr[kd,bf[kf]:bf[kf+1]-1] = ( cc LT maxint ) ? indgen( cc ) : $
		[ indgen( maxint ), intarr( cc - maxint ) + maxint ]
      endfor
    ENDFOR

;;; set the glitch flags
    status = tsd->set_pix_flag_tsd( 'fis_obs', 'gpgl_type1', byte( flag EQ 1 ))
    status = tsd->set_pix_flag_tsd( 'fis_obs', 'gpgl_type2', byte( flag EQ 2 ))
    status = tsd->set_pix_flag_tsd( 'fis_obs', 'gpgl_type3', byte( flag EQ 3 ))
    status = tsd->set_pix_flag_tsd( 'fis_obs', 'gpgl_type4', byte( flag EQ 4 ))
    status = tsd->set_pix_flag_tsd( 'fis_obs', 'gpgl_tail',  byte( flag EQ 5 ))
    status = tsd->set_pix_flag_tsd( 'fis_obs', 'no_gpgl',    byte( flag EQ 9 ))
    status = tsd->set_quality_tsd( 'fis_obs', 'qual_gpgl_corr',  qual )
    badf = tsd->get_pix_flag_tsd( 'fis_obs', 'bad' ) OR $
			byte( flag GT 0 AND flag LE 5 )
    status = tsd->set_pix_flag_tsd( 'fis_obs', 'bad', badf )
    status = tsd->set_value_tsd( 'fis_obs', 'cnt_glitch_gp', cntr )

    if ( n_elements( glitchlist ) GT 0 ) then free_lun, glunit
    if ( pltype EQ 7 ) then stop_hc

;;; Write HISTORY in header ;;;

    lmessage, 7, 'Writing HISTORY',/CONTINUE
    hist_message = ' gb_set_glitch_status.pro,v 0.0 done. CCF= '

    IF KEYWORD_SET(ccf) NE 1 THEN BEGIN
	hist_message = hist_message + 'NONE'
    ENDIF ELSE BEGIN
	hist_message = hist_message + ccf
    ENDELSE

    dummy = tsd_fxaddhist(tsd, hist_message)
    IF dummy NE 0 THEN $
	MESSAGE,'Unable to write HISTORY',/CONTINUE

;    print, systime(1) - syst

    return, !SUCCEED

END


