;#> bayesfit.dc2
; Identifier  bayesfit
;
; Purpose     Calculate the solution + evidence for the matrix + vector
;
; Synopsis    error = bayesfit( [matrix,][vector,] sumlogsigma, ndata, $
;			chisq=chisq, sol=sol, stdev=stdev, scale=scale,$
;			logpr=logpr, covar=covar, conformal=conformal )
;
; Arguments	Name	I/O	Type:	 Description
;		-------------------------------------------------------------
;		matrix	 I	fltarr	 curvature matrix
;			 if not present  logpr is calculated assuming there
;					 are no parameters in the model
;		vector	 I	fltarr	 pertaining vector
;			 if not present  no solution is calculated and
;					 chisq remains unchanged.
;					 To be used for non-lin problems
;		slsigma  I	float	 sum of log10( stdev( data ) )
;		ndata	 I	integer  number of datapoints
;		chisq	 I	float	 sum over the dependent variable
;			 O		 chisq of the fit
;		sol	 I	fltarr	 prior values of params (only if conformal)
;			 O		 if vector present: the solution found
;		stdev	 I	fltarr	 prior stdev of the parameters
;			 O		 the standard deviations found
;		scale	 I	float	 0  : perform automatic scaling
;					 >0 : factor to scale all errors with
;					 <0 : automatic scaling with minimum
;					      minimum = abs( scale )
;					 default: 1.0
;			 O	float	 estimated scaling factor of sigma
;		covar	 O	fltarr	 covariance matrix = chisq*inverse(mat)
;		logprob  O	float	 log10( evidence )
;		conformal I	keyword  use conformal (ie. gaussian) priors
;					 otherwise use uniform priors
;
; Returns:	error	 0	int	 0 : OK
;					 1 : matrix or vector is infinite or NaN
;					 2 : det(matrix) < 0
;					 3 : chisq < 0
;					 4 : improper # of params in call
;					 5 : degenerate matrix
;
; Description:	Crouts method for solving normal equations. And then MacKay's
;		method is used to calculate the evidence.
;
;		When the problem is LINEAR in its parameters the procedure can
;		be used to FIND the solution. Otherwise other methods (eg.
;		simplex, amoebe, gradient descent) must be used to get at the
;		optimal solution.
;		In both cases the evidence can be calculated from the curvature
;		matrix.
;
;
; Comment:	Key ideas taken from David McKay (1992) in
;		"Maximum Entropy and Bayesian Methods" Eds C.R. Smith etal.
;		Kluwer Academic Publishers
;
;
; Example	--
;
; Category	UTIL
;
; Filename	bayesfit.pro
;
; Author	Do Kester
;
; Version	4.5
;
; History:	0.0 24/10/95	DK C version
;		1.0 11/02/97	DK IDL version
;		2.0 12/11/98	DK return error in stead of logprob
;		3.0 17/08/99	DK add covariance matrix option
;		4.0 02/11/99	DK make matrix and vector optional
;		4.1 16/12/99	DK automatic noise scaling
;		4.2 09/03/00	DK adapt to pipeline
;		4.3 15/03/01	DK change only internal variables
;		4.4 03/04/01	DK remove compile_opt
;		4.5 21/05/01	DK default for scale + doc update
;#<

FUNCTION bayesfit, matrix, vector, sumlogsigma, ndata, chisq=chisq, sol=sol, $
		   stdev=stdev, scale=scale, logpr=logprob, $
		   covar=covar, conformal=conformal

; COMPILE_OPT IDL2

FORWARD_FUNCTION lu_chisq, lu_stdev, lu_logdet, lusol

CASE n_params() OF
  4: BEGIN				    ; all params present -> do all
       np = n_elements( vector )
       slsig = sumlogsigma
       ndat = ndata
       vec = [vector]			    ; do not overwrite vector
       want_sol = 1
     END
  3: BEGIN				    ; no vector -> no solution
       ndat = sumlogsigma		    ; shift ndata
       slsig = vector			    ; shift sumlogsigma
       s = size( matrix )
       np = s[1]
       vec = make_array( np, type=s[3] )
       want_sol = 0
     END
  2: BEGIN				    ; no matrix, no vector ->no params
       slsig = matrix			    ; shift ndata
       ndat = vector			    ; shift sumlogsigma
       logoccam = 0
       np = 0
       logdet = 0

       goto, calc_logpr 	    ; skip all other stuff: no params at all

     END
  ELSE: BEGIN
	lmessage, 9, /continue, $
	    string( 'Improper number of parameters in call:', n_params() )
	return, 4
     END
ENDCASE

mat = reform( [matrix], np, np )	    ; keep as matrix in case of length 1
b = where( finite( [ reform( mat, np*np ), vec ] ) EQ 0, nb )
IF ( nb GT 0 ) THEN BEGIN
  lmessage, 9, "NaNs in matrix and/or vector", /continue
  RETURN, 1
ENDIF

IF ( n_elements( sol )	 NE np ) THEN sol   = make_array( size=size( vec ) )
IF ( n_elements( stdev ) LT np ) THEN $
	    stdev = make_array( size=size( vec ), value=1 )

b = indgen( np )
IF ( keyword_set( conformal ) ) THEN BEGIN
  wgt = 1.0 / ( stdev[b] * stdev[b] )
  mat[b,b] = mat[b,b] + wgt			; add prior information to mat
  chisq    = chisq  + total( wgt )		; extra `observations' to syy
  IF ( want_sol ) THEN BEGIN
    sol[b]   = sol[b] * wgt
    vec[b]   = vec[b] + sol[b]			; add prior information to vec
  ENDIF
ENDIF

; for conformal priors the following should be slightly different:
; something like the value of the gaussian prior at the true solution
; for now the value for a uniform prior is good enough
logoccam = - total( alog10( stdev[b] ) )

; perform LU decomposition on the matrix
if ( ispresent( total( mat, 1 ) EQ 0 ) ) then begin
  lmessage, 9, "Degenerate matrix", /continue
  return, 5
end
ludc, mat, index, interchanges=parity
mat = reform( mat, np, np )		; keep as matrix in case of length 1

; make a new solution by backsubstitution
IF ( want_sol ) THEN BEGIN
    sol = lusol( mat, index, vec )
    chisq = lu_chisq( chisq, vec, sol )
ENDIF

IF ( chisq LT 0 ) THEN BEGIN
  lmessage, 9, 'Negative chisq: should not happen.', /continue
  RETURN, 3
ENDIF

;print, mat, index, chisq, covar, stdev
;print, b

; determine the standard deviations
;print, lu_stdev( mat, index, chisq, covar=covar )
stdev[b] = lu_stdev( mat, index, chisq, covar=covar )

;print, stdev

; det( mat ) == inverse( det( inverse( mat ) ) )
logdet = lu_logdet( mat, parity=parity )
IF ( parity LE 0 ) THEN BEGIN
  lmessage, 9, string( 'Determinant < 0 ', -( 10 ^ logdet ), $
			' : should not happen' ), /continue
  RETURN, 2
ENDIF

calc_logpr:					; calculate logprob

dof = ndat - np

if ( n_elements( scale ) EQ 0 ) then scale = 1.0
; autoscaling
IF ( scale GT 0 ) THEN s2 = 1.0 / scale^2 $
ELSE BEGIN
  if ( scale LT 0 ) then s2 = ( chisq + ndat / scale^2 ) / dof $
  else s2 = chisq / dof
  logoccam = logoccam + alog10( sqrt( !PI / ndat ) / stdev[np] )
ENDELSE

; get Occam-factor
logoccam = logoccam + 0.5 * ( np * alog10( 2 * !PI * s2 ) + logdet )

; calculate scale factor
scale = sqrt( chisq / ( ndat - np ) )

; get likelihood
loglikelihood = - 0.5 * chisq / s2 * alog10( exp( 1.0 ) ) $
			- slsig - ndat * alog10( sqrt( 2 * !PI * s2 ) )
; and the evidence
logprob = loglikelihood + logoccam

;print, ndat, np, logdet, logoccam, chisq, scale, loglikelihood, $
;	logprob, format='(2i4,8f10.4)'

RETURN, 0

END
