;+
; NAME:
;	im_detcoad_pointdrizzle.pro
; PURPOSE:
;	coadd one detector scan into the mapset struct
; CATEGORY:
;	Function.
; CALLING SEQUENCE:
;	Result = im_detcoad_pointdrizzle( flux, x, y, mapset, mapinfo, TBD )
;
; INPUTS:
;	flux		valid flux for this det
;	x, y		pertaining projected coordinates in pixel values
;	mapinfo 	map information
; OUTPUTS:
;	mapset		update of contents
; RETURN
;	!SUCCEED if function ended successfully.
;	!FAIL in failure.
; RESTRICTIONS:
;
; PROCEDURE:
;	Add flux to the pixel nearest to the data sample position *ONLY*
;	(This mimics point drizzle map-making; it keeps the noise correlation structure simpler!)
;	Increment the same pixels in the count-map and add the square
;	of the fluxes into the pixels of the stdev-map
;
; EXAMPLE:
;
; REFERENCE:
;	Baba et al. 2002 Nov. 21, ``ASTRO-F/FIS Survey Data Structure"
;
; MODIFICATION HISTORY:
;
;	Thu Nov 24 09:13:00 2005, Richard Savage
;	<rss23@iris70.ir.isas.jaxa.jp>
;
;		- Created.
;-
FUNCTION im_detcoad_pointdrizzle, flux, wgt, xx, yy, mapset, mapinfo

;;----------------------------------------------------------------------
;; DO'S SETTINGS -------------------------------------------------------
;;----------------------------------------------------------------------
COMPILE_OPT IDL2
;ON_ERROR 2
;;----------------------------------------------------------------------
;; FIND THE NEAREST PIXELS VALUES FOR EACH DATA SAMPLE -----------------
;;----------------------------------------------------------------------
roundedX = round(xx)
roundedY = round(yy)
;;----------------------------------------------------------------------
;; IDENTIFY THE SAMPLES THAT ACTUALLY FALL WITHIN THE IMAGE ------------
;;----------------------------------------------------------------------
valid = where(roundedX GE 0		    AND $
	      roundedX LT mapinfo.pixels[0] AND $
	      roundedY GE 0		    AND $
	      roundedY LT mapinfo.pixels[1],	$
	      nValid)
;;----------------------------------------------------------------------
;; UPDATE MAPSET -------------------------------------------------------
;;----------------------------------------------------------------------
;;note that for ease, I've kept this as a loop.  You can use array arithmetic
;;for this part, but you have to be careful about duplicate pixel positions
;;(ask me if you really care about this :-) )
IF (nValid GT 0) THEN BEGIN
    ;;for convenience, separate out the valid points
    rx	  = roundedX[valid]
    ry	  = roundedY[valid]
    rFlux = flux[valid]
    rWgt  = wgt[valid]
    ;;update the mapset information
    FOR i=0l, nValid-1 DO BEGIN
	mapset.count[rx[i], ry[i]] += 1.0
	mapset.mom0[rx[i], ry[i]] += rWgt[i]
	mapset.mom1[rx[i], ry[i]] += rWgt[i] * rFlux[i]
	mapset.mom2[rx[i], ry[i]] += rWgt[i] * rFlux[i] * rFlux[i]
    ENDFOR
ENDIF
;;----------------------------------------------------------------------
;; RETURN --------------------------------------------------------------
;;----------------------------------------------------------------------
return, !SUCCEED


END
;*****************************************************************************
;*****************************************************************************
;;----------------------------------------------------------------------
;; ----------------------------------------
;;----------------------------------------------------------------------

