;+
; NAME:
;   ss_coad_image
;
; PURPOSE:
;   Make co-add image of the processed data
;
; OUTPUT:
;    Co-add image data, RADEC (IDL): '*_img.sav'
;    Co-add image data, RADEC, and parameters (IDL): '*_img_[w|n].fits'
;
; CALLING SEQUENCE:
;
; MODIFICATION HISTORY:
;   Written by S. Matsuura, 2006/09/05
;   2006/10/04  S.Matsuura,   minor modification
;   2006/10/18  I.Yamamura,   renamed as ss_coad_image
;   2006/11/28  S.Matsuura,   save coad image data in IDL format
;   2006/12/21  I.Yamamura,   exit gently when no GADS is available
;   2006/12/21  I.Yamamura,   window control revised not to use too many windows
;   2006/12/26  S.Matsuura,   include 'AOT' as an input parameter
;   2006/12/27  S.Matsuura,   changed ra-dec grid size, include 'RST' as an input parameter
;   2006/12/29  Makiuti, S.,  Added /NO_DISPLAY options.
;   2006/12/29  S.Matsuura,   included area boundary information as input parameter.
;                             In case of "MAPPING" this boundary determines the grid in entire map.
;   2006/12/30  M.Shirahata,  modify the plot style.
;   2007/01/05  Makiuti, S.,  Improved error handling. 
;                             Added ERROR_STATUS option argument to return error status.
;   2007/01/06  I.Yamamura,   Data over R.A.=360deg can be mapped.
;   2007/01/09  S.Matsuura,   Sigma for bad data rejection can be specified.
;   2007/01/10  I.Yamamura,   image plot in right R.A. orientation and square grid.
;   2007/01/10  S.Matsuura,   chenged position of IF-statement for "mapping".
;   2007/01/17  I.Yamamura,   /ISOTROPIC option removed from CONTOUR call.
;   2007/01/18  M.Shirahata,  changed the default value of sigma from 5 to 2.
;   2007/01/18  S.Matsuura,   changed the scan speed limit to 30"/s for special operation.
;   2007/01/20  S.Matsuura,   simple mean -> weighted mean for co-adding.
;   2007/01/23  S.Matsuura,   included bad data flag found in the flat fielding process.
;   2007/01/30  S.Matsuura,   if det=0 then no contribution to coad image.
;   2007/01/30  M.Shirahata,  change how to treat the t_start and t_end options.
;                             add new option (/newmapping method).
;                             abolished making the PS file.
;   2007/02/02  M.Shirahata,  add data_type in output IDL format data.
;   2007/02/05  S.Matsuura,   small modification in area boundary determination process.
;   2007/02/06  S.Matsuura,   grid size is autonatically set according to N_RAMP_DIV(n_int).
;   2007/02/06  M.Shirahata,  add /SCUT option.
;   2007/02/06  S.Matsuura,   bug fix for add /SCUT option.
;   2007/02/06  M.Shirahata,  return to previous version & bug fix for /SCUT option.
;   2007/02/07  M.Shirahata,  bug fix.
;   2007/02/08  S.Matsuura,   can specify grid size as input option.
;                 Auto grid sizing with n_int is disused.
;   2007/02/08  S.Matsuura,   calculation of RMS fluctution of final coadded data.
;   2007/02/09  S.Matsuura,   Added calculation of mean level of coad image and revised RMS fluctution.
;   2007/02/09  S.Matsuura,   Included map_mean, dec_grid and ra_grid in saved IDL-format image data. 
;   2007/02/21  Makiuti, S.,  Set a color table explicitly for contour map display.
;   2007/03/27  M.Shirahata,  Small bug fix for IDL plot. No change to data analysis process.
;   2007/06/06  M.Shirahata,  Small bug fix. 
;
;-
pro ss_coad_image, x, ra, dec, det, err, eg_gl_cnt, $
                   file_header, data_type, AOT, RST, rst_int, n_int, bad_flg, $
                   im_ra, im_dec, coad_det, coad_num, coad_err, $
                   ra_min, ra_max, dec_min, dec_max, $
                   NO_DISPLAY = no_display, ERROR_STATUS = error_status, $
                   SIGMA = sigma, T_START = t_start, T_END = t_end, $ 
                   MAPPING = mapping, NEWMAPPING = newmapping, SCUT = scut, $
           GRID_SW = grid_sw, GRID_LW = grid_lw

!p.multi = [0,0,0]
RAD2SEC = 206264.8062470963551572D ;[arcsec/rad]
RAD2DEG = 57.295779513082320877D   ;[deg/rad]


; Check the parameters and initialize
IF (n_elements(no_display) EQ 0) THEN no_display=0


; variables specified for detector
if(data_type eq 'FIS_SW') then begin
    c_pix = 20
    n_pix = 100
    w_pix = 60
    beam_size = 40.0D         ; FWHM in 1-D profile
    smpl_freq = 25.28D        ; [Hz]
endif else begin   ;'FIS_LW'
    c_pix = 15
    n_pix = 75
    w_pix = 45
    beam_size = 60.0D         ; FWHM in 1-D profile
    smpl_freq = 16.86D        ; [Hz]
endelse


if keyword_set(mapping) then goto, skipmap1
if keyword_set(newmapping) then goto, skipnewmap1

; Scan Cut
if (scut eq 0) then goto, skipscut1

; in case of scut=ON
x0=x
ra0=ra
dec0=dec
det0=det
err0=err
eg_gl_cnt0=eg_gl_cnt

CASE AOT OF
    'PV_FIS01':       scan=4
    'PV_FIS02':       scan=2
    'P12_FIS01':      scan=4
    'P12_FIS02':      scan=2
    'P12_FIS01_SAA':  scan=4
    'P12_FIS02_SAA':  scan=2
ENDCASE

; Scan Cut Loop Start
for sloop=0, scan DO BEGIN
    
; File name etc
    snum = Strcompress(String(sloop+1), /remove_all)
    x=x0
    ra=ra0
    dec=dec0
    det=det0
    err=err0
    eg_gl_cnt=eg_gl_cnt0

    ; Time range set
    if (AOT eq 'PV_FIS01') then begin
        if (sloop eq 0) then begin
            t_start=750
            t_end=900
        endif
        if (sloop eq 1) then begin
            t_start=926
            t_end=1050
        endif
        if (sloop eq 2) then begin
            t_start=1076
            t_end=1200
        endif
        if (sloop eq 3) then begin
            t_start=1226
            t_end=1350
        endif
        if (sloop eq scan) then begin
            t_start=750
            t_end=1350
        endif
    endif
        
    if (AOT eq 'PV_FIS02') then begin
        if (sloop eq 0) then begin
            t_start=750
            t_end=1050
        endif
        if (sloop eq 1) then begin
            t_start=1076
            t_end=1350
        endif
        if (sloop eq scan) then begin
            t_start=750
            t_end=1350
        endif
    endif
    
    if ((AOT eq 'P12_FIS01') || (AOT eq 'P12_FIS01_SAA')) then begin
        if (sloop eq 0) then begin
            t_start=630
            t_end=787
        endif
        if (sloop eq 1) then begin
            t_start=818
            t_end=975
        endif
        if (sloop eq 2) then begin
            t_start=1005
            t_end=1162
        endif
        if (sloop eq 3) then begin
            t_start=1193
            t_end=1350
        endif
        if (sloop eq scan) then begin
            t_start=630
            t_end=1350
        endif
    endif

    if ((AOT eq 'P12_FIS02') || (AOT eq 'P12_FIS02_SAA')) then begin
        if (sloop eq 0) then begin
            t_start=630
            t_end=975
        endif
        if (sloop eq 1) then begin
            t_start=1005
            t_end=1350
        endif
        if (sloop eq scan) then begin
            t_start=630
            t_end=1350
        endif
    endif

    print, 'Time range was set from', Strcompress(t_start) +' to', Strcompress(t_end) +'.'

    ; t_start
    t_start_idx = where(x gt t_start*smpl_freq, t_start_cnt)
    x = x[t_start_idx]
    n_line = n_elements(x)
    ra = ra[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]
    dec = dec[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]
    det = det[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]
    err = err[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]
    eg_gl_cnt = eg_gl_cnt[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]

    ; t_end
    t_end_idx = where(x lt t_end*smpl_freq, t_end_cnt)
    x = x[t_end_idx]
    n_line = n_elements(x)
    ra = ra[*,0:t_end_cnt-1]
    dec = dec[*,0:t_end_cnt-1]
    det = det[*,0:t_end_cnt-1]
    err = err[*,0:t_end_cnt-1]
    eg_gl_cnt = eg_gl_cnt[*,0:t_end_cnt-1]

skipscut1:

if (scut ne 0) then goto, skipcut2

; Time range set
if (n_elements(t_start) eq 0) then begin
    CASE AOT OF
        'PV_FIS01':       t_start=750
        'P12_FIS01':      t_start=630
        'P12_FIS01_SAA':  t_start=630
        'PV_FIS02':       t_start=750
        'P12_FIS02':      t_start=630
        'P12_FIS02_SAA':  t_start=630
    ENDCASE
endif
if (n_elements(t_end) eq 0) then begin
    t_end = 1350
endif
print, 'Time range was set from', Strcompress(t_start) +' to', Strcompress(t_end) +'.'

; t_start
t_start_idx = where(x gt t_start*smpl_freq, t_start_cnt)
x = x[t_start_idx]
n_line = n_elements(x)
ra = ra[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]
dec = dec[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]
det = det[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]
err = err[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]
eg_gl_cnt = eg_gl_cnt[*,t_start_idx[0]:t_start_idx[0]+t_start_cnt-1]

; t_end
t_end_idx = where(x lt t_end*smpl_freq, t_end_cnt)
x = x[t_end_idx]
n_line = n_elements(x)
ra = ra[*,0:t_end_cnt-1]
dec = dec[*,0:t_end_cnt-1]
det = det[*,0:t_end_cnt-1]
err = err[*,0:t_end_cnt-1]
eg_gl_cnt = eg_gl_cnt[*,0:t_end_cnt-1]

skipcut2:
skipnewmap1:


; R.A. in relative expression
    ra_med = MEDIAN(ra)
    print, 'RA_mid = ', ra_med
    ra -= ra_med
    if ((MAX(ra) - MIN(ra)) GT 300D) then begin
        if (ra_med GT 180D) then begin
            idx = WHERE (ra LT -180D, cnt)
            if (cnt GT 0) then ra[idx] += 360D
        endif else begin
            idx = WHERE (ra GT 180D, cnt)
            if (cnt GT 0) then ra[idx] -= 360D
        endelse
    endif

    ; scan-speed calculation
    n_int_dbl = double(n_int)
    data_int = rst_int/n_int_dbl
    print, 'Data interval [s]:', data_int
    if(data_int gt 0) then begin
        sp_dec = (dec[0,*]-shift(dec[0,*], 1))/data_int*3600.0D
        sp_ra = cos(dec[0,*]/RAD2DEG)*(ra[0,*]-shift(ra[0,*], 1))/data_int*3600.0D
        abs_speed = sqrt(sp_dec^2+sp_ra^2)
        med_sp_dec = median(abs(sp_dec))
        med_sp_ra = median(abs(sp_ra))
        med_speed = median(abs_speed)
               
        print, 'Median scan speed [arcsec/s]'
        print, '    angular_speed(RA):', med_sp_ra
        print, '    angular_speed(DEC):', med_sp_dec
        print, '    Absolute:', med_speed
    endif else begin
        message, 'maybe CDS mode, but... Check data!', /continue
        error_status = 1 ; set error status
        return
    endelse
    if (med_speed gt 3 and med_speed lt 5) then scan_speed = 4
    if (med_speed gt 7 and med_speed lt 9) then scan_speed = 8
    if (med_speed gt 14 and med_speed lt 16) then scan_speed = 15
    if (med_speed gt 29 and med_speed lt 31) then scan_speed = 30
    print, 'AOT specified scan speed [arcsec/s]: ', scan_speed
    if (med_speed lt 3 and med_speed gt 31) then begin
    MESSAGE, "Unknown scan speed. Process stopped. Check data!", /CONTINUE
    error_status = 1 ; set error status
    return
    endif


    ; define scan area boundary
    scan_t_idx = where(abs_speed lt 31, scan_t_cnt)
        if (scan_t_cnt GT 0) then begin
            ra_sub = ra[*,scan_t_idx]
            dec_sub = dec[*,scan_t_idx]
        endif else begin
            MESSAGE, "No proper attitude information. Process stopped. Check data!", /CONTINUE
            error_status = 1 ; set error status
            return
        endelse

    ra_max = max(ra_sub)
    ra_min = min(ra_sub)
    dec_max = max(dec_sub)
    dec_min = min(dec_sub)

skipmap1:

    print, 'Plot area [deg]'
    print, '    RA_min:', ra_min
    print, '    RA_max:', ra_max
    print, '    DEC_min:', dec_min
    print, '    DEC_max:', dec_max

    ra_mean = (ra_max + ra_min)/2.0D
    dec_mean = (dec_max + dec_min)/2.0D
    print, 'area center (RA, DEC): ', ra_mean, dec_mean


; grid in RADEC
 
    if (AOT eq 'PV_FIS01' or AOT eq 'P12_FIS01' or AOT eq 'P12_FIS01_SAA') then begin
    if(data_type eq 'FIS_SW') then begin
        grid_size = 15D/3600D
        if (n_elements(grid_sw) eq 1) then grid_size = grid_sw/3600D
    endif else begin
        grid_size = 30D/3600D
        if (n_elements(grid_lw) eq 1) then grid_size = grid_lw/3600D
    endelse
    grid_limit = beam_size/3600D/n_int*(scan_speed/15D)/4D
    endif else begin
    if(data_type eq 'FIS_SW') then begin
        grid_size = 30D/3600D
        if (n_elements(grid_sw) eq 1) then grid_size = grid_sw/3600D
    endif else begin
        grid_size = 60D/3600D
        if (n_elements(grid_lw) eq 1) then grid_size = grid_lw/3600D
    endelse
    grid_limit = beam_size/3600D/n_int*(scan_speed/15D)/2D
    endelse

    print, 'Grid size is set to [arcsec]: ', grid_size*3600D

    ; auto limit of grid size
    if(grid_size lt grid_limit) then begin
    print, '% Input grid size is too small! It may cause a lot of empty grid cells in the map.'
    grid_size = grid_limit
    print, 'Grid size is automatically set to [arcsec]: ', grid_size*3600D
    endif

    dec_grid = grid_size
    ra_grid = grid_size/cos(dec_mean/RAD2DEG)

    print, 'grid size taken for calculation:'
    print, '    RA [arcsec] :', ra_grid*3600.0D
    print, '    DEC [arcsec]:', dec_grid*3600.0D

    n_data = n_elements(det)/n_pix

    ra_num = fix((ra_max-ra_min)/ra_grid) + 1
    dec_num =  fix((dec_max-dec_min)/dec_grid) + 1


; initial co-add calculation

    print, 'making initial co-add image ... '

    im_ra = dblarr(ra_num)
    im_dec = dblarr(dec_num)
    coad_num0 = dblarr(2, ra_num, dec_num)
    coad_det0 = dblarr(2, ra_num, dec_num)
    coad_err0 = dblarr(2, ra_num, dec_num)
    coad_weight0 = dblarr(2, ra_num, dec_num)
    coad_mean0 = dblarr(2, ra_num, dec_num)
    coad_det_sq0 = dblarr(2, ra_num, dec_num)
    coad_stdev0 = dblarr(2, ra_num, dec_num)

    ra_idx = dindgen(ra_num)
    dec_idx = dindgen(dec_num)
    im_ra = ra_min + ra_idx*ra_grid
    im_dec = dec_min + dec_idx*dec_grid

    for pix = 0, w_pix-1 do begin
    if(bad_flg[pix] eq 0) then begin
        for j = 0L, n_data-1 do begin
            i_ra = fix((ra[pix,j]-ra_min)/ra_grid)
            i_dec = fix((dec[pix,j]-dec_min)/dec_grid)
            if((i_ra ge 0 and i_dec ge 0) $
               and (i_ra lt ra_num and i_dec lt dec_num) $
               and (eg_gl_cnt[pix,j] eq 0) $
               and (det[pix,j] ne 0D) $
           and (err[pix,j] gt 0D) ) then begin
                coad_num0[0,i_ra,i_dec] = coad_num0[0,i_ra,i_dec] + 1D
            coad_det0[0,i_ra,i_dec] = coad_det0[0,i_ra,i_dec] + det[pix,j]/err[pix,j]^2
                coad_weight0[0,i_ra,i_dec] = coad_weight0[0,i_ra,i_dec] + 1D/err[pix,j]^2
                coad_mean0[0,i_ra,i_dec] = coad_mean0[0,i_ra,i_dec] + det[pix,j]
                coad_det_sq0[0,i_ra,i_dec] = coad_det_sq0[0,i_ra,i_dec] + det[pix,j]^2
            endif
        endfor
    endif
    endfor

    for pix = w_pix, n_pix-1 do begin
    if(bad_flg[pix] eq 0) then begin
        for j = 0L, n_data-1 do begin
            i_ra = fix((ra[pix,j]-ra_min)/ra_grid)
            i_dec = fix((dec[pix,j]-dec_min)/dec_grid)
            if( (i_ra ge 0 and i_dec ge 0) $
               and (i_ra lt ra_num and i_dec lt dec_num) $
               and (eg_gl_cnt[pix,j] eq 0) $
               and (det[pix,j] ne 0D) $
           and (err[pix,j] gt 0D) ) then begin
                coad_num0[1,i_ra,i_dec] = coad_num0[1,i_ra,i_dec] + 1D
                coad_det0[1,i_ra,i_dec] = coad_det0[1,i_ra,i_dec] + det[pix,j]/err[pix,j]^2
                coad_weight0[1,i_ra,i_dec] = coad_weight0[1,i_ra,i_dec] + 1D/err[pix,j]^2
                coad_mean0[1,i_ra,i_dec] = coad_mean0[1,i_ra,i_dec] + det[pix,j]
                coad_det_sq0[1,i_ra,i_dec] = coad_det_sq0[1,i_ra,i_dec] + det[pix,j]^2
            endif
        endfor
    endif
    endfor


; calculate mean and std_dev

    nz_idx = where(coad_num0 gt 0 and coad_weight0 gt 0, nz_cnt)
    print, 'Non-zero co-add number: ', nz_cnt
    if(nz_cnt ne 0) then begin
        coad_det0[nz_idx] = coad_det0[nz_idx]/coad_weight0[nz_idx]
        coad_err0[nz_idx] = sqrt(1D/coad_weight0[nz_idx])
        coad_mean0[nz_idx] = coad_mean0[nz_idx]/coad_num0[nz_idx]
        coad_det_sq0[nz_idx] = coad_det_sq0[nz_idx]/coad_num0[nz_idx]
        coad_stdev0[nz_idx] = sqrt(coad_det_sq0[nz_idx]-coad_mean0[nz_idx]^2)
    endif


; remake co-add image with 2-sigma rejection

    coad_num = dblarr(2, ra_num, dec_num)
    coad_det = dblarr(2, ra_num, dec_num)
    coad_err = dblarr(2, ra_num, dec_num)
    coad_weight = dblarr(2, ra_num, dec_num)
    coad_mean = dblarr(2, ra_num, dec_num)
    coad_det_sq = dblarr(2, ra_num, dec_num)
    coad_stdev = dblarr(2, ra_num, dec_num)

    if n_elements(sigma) eq 0 then sigma=2D
    print, 're-producing image with sigma rejection ...'
    print, 'Sigma : ', sigma

    for pix = 0, w_pix-1 do begin
    if(bad_flg[pix] eq 0) then begin
        for j = 0L, n_data-1 do begin
            i_ra = fix((ra[pix,j]-ra_min)/ra_grid)
            i_dec = fix((dec[pix,j]-dec_min)/dec_grid)
            if((i_ra ge 0 and i_dec ge 0) $
               and (i_ra lt ra_num and i_dec lt dec_num) $
               and (eg_gl_cnt[pix,j] eq 0) $
               and (det[pix,j] ne 0D) $
           and (err[pix,j] gt 0D) ) then begin
                if(abs(det[pix,j]-coad_det0[0,i_ra,i_dec]) lt sigma*coad_stdev0[0,i_ra,i_dec]) then begin $
                coad_num[0,i_ra,i_dec] = coad_num[0,i_ra,i_dec] + 1D
                coad_det[0,i_ra,i_dec] = coad_det[0,i_ra,i_dec] + det[pix,j]/err[pix,j]^2
                coad_weight[0,i_ra,i_dec] = coad_weight[0,i_ra,i_dec] + 1D/err[pix,j]^2
                coad_mean[0,i_ra,i_dec] = coad_mean[0,i_ra,i_dec] + det[pix,j]
                coad_det_sq[0,i_ra,i_dec] = coad_det_sq[0,i_ra,i_dec] + det[pix,j]^2
                endif
            endif
        endfor
    endif
    endfor

    for pix = w_pix, n_pix-1 do begin
    if(bad_flg[pix] eq 0) then begin
        for j = 0L, n_data-1 do begin
            i_ra = fix((ra[pix,j]-ra_min)/ra_grid)
            i_dec = fix((dec[pix,j]-dec_min)/dec_grid)
            if((i_ra ge 0 and i_dec ge 0) $
               and (i_ra lt ra_num and i_dec lt dec_num) $
               and (eg_gl_cnt[pix,j] eq 0) $
               and (det[pix,j] ne 0D) $
           and (err[pix,j] gt 0D) ) then begin
                if(abs(det[pix,j]-coad_det0[1,i_ra,i_dec]) lt sigma*coad_stdev0[1,i_ra,i_dec]) then begin $
                coad_num[1,i_ra,i_dec] = coad_num[1,i_ra,i_dec] + 1D
                coad_det[1,i_ra,i_dec] = coad_det[1,i_ra,i_dec] + det[pix,j]/err[pix,j]^2
                coad_weight[1,i_ra,i_dec] = coad_weight[1,i_ra,i_dec] + 1D/err[pix,j]^2
                coad_mean[1,i_ra,i_dec] = coad_mean[1,i_ra,i_dec] + det[pix,j]
                coad_det_sq[1,i_ra,i_dec] = coad_det_sq[1,i_ra,i_dec] + det[pix,j]^2
                endif
            endif
        endfor
    endif
    endfor


; re-calculate mean and std_dev

    nz_idx = where(coad_num gt 0 and coad_weight gt 0, nz_cnt)
    print, 'Non-zero co-add number: ', nz_cnt
    if(nz_cnt ne 0) then begin  
        coad_det[nz_idx] = coad_det[nz_idx]/coad_weight[nz_idx]
        coad_err[nz_idx] = sqrt(1D/coad_weight[nz_idx])
        coad_mean[nz_idx] = coad_mean[nz_idx]/coad_num0[nz_idx]
        coad_det_sq[nz_idx] = coad_det_sq[nz_idx]/coad_num[nz_idx]
        coad_stdev[nz_idx] = sqrt(coad_det_sq[nz_idx]-coad_mean[nz_idx]^2)
    endif


; cell-to-cell rms noise in the coadd map

    map_mean = dblarr(2,4)
    for i = 0, 1 do begin
    p_coad_det = dblarr(ra_num,dec_num)
    p_coad_det[*,*] = coad_det[i,*,*]
    p_nz_idx = where(p_coad_det ne 0, p_nz_cnt)
    delta = 1
    cnt = 0
    if (p_nz_cnt gt 0) then begin
        p_mean = mean(p_coad_det[p_nz_idx],/double)
        p_rms = stddev(p_coad_det[p_nz_idx],/double)
        while(delta gt 0.05 or cnt lt 5) do begin
        p_mean0 = p_mean
        p_rms0 = p_rms
        nz_idx = where(abs(p_coad_det[p_nz_idx]-p_mean0) lt 3*p_rms0, nz_cnt)
        p_nz_idx = p_nz_idx[nz_idx]
        if (nz_cnt gt 0) then begin
            p_mean = mean(p_coad_det[p_nz_idx],/double)
            p_rms = stddev(p_coad_det[p_nz_idx],/double)
        endif
        delta = abs(p_rms-p_rms0)/p_rms0
        cnt += 1
        endwhile
    endif
    map_mean[i,0] = p_mean
    map_mean[i,1] = p_rms
    map_mean[i,2] = p_nz_cnt
    map_mean[i,3] = p_nz_cnt-nz_cnt
    endfor

    print, 'Mean level of coad image and RMS fluctuation after 3-sigma rejection [MJy/sr]:'
    print, '    WIDE:   ', map_mean[0,0], map_mean[0,1]
    print, '    NARROW: ', map_mean[1,0], map_mean[1,1]
    print, 'Number of cells with non-zero values and 3-sigma rejected cells:'
    print, '    WIDE:   ', map_mean[0,2], map_mean[0,3]
    print, '    NARROW: ', map_mean[1,2], map_mean[1,3]


if keyword_set(mapping) then goto, skipmap2

; R.A. from relative to absolute

   ra += ra_med
   im_ra += ra_med

; calculate max and min values

    coad_wide_max = max(coad_det[0,*,*], wide_max_sub, min=coad_wide_min)
    print, 'max_value in wide coad-map [MJy/sr] and max_idx:', coad_wide_max, wide_max_sub
    print, 'min_value in wide coad-map [MJy/sr]:', coad_wide_min
    wide_max_str = string(coad_wide_max)
    wide_max_str = strtrim(wide_max_str, 2)
    wide_min_str = string(coad_wide_min)
    wide_min_str = strtrim(wide_min_str, 2)

    coad_narrow_max = max(coad_det[1,*,*], narrow_max_sub, min=coad_narrow_min)
    print, 'max_value in narrow coad-map [MJy/sr] and max_idx:', coad_narrow_max, narrow_max_sub
    print, 'min_value in narrow coad-map [MJy/sr]:', coad_narrow_min
    narrow_max_str = string(coad_narrow_max)
    narrow_max_str = strtrim(narrow_max_str, 2)
    narrow_min_str = string(coad_narrow_min)
    narrow_min_str = strtrim(narrow_min_str, 2)

    grid_str = string(dec_grid*3600.0)
    grid_str = strtrim(grid_str, 2)


; plot contour

    help, im_ra, im_dec, coad_det

    IF (no_display NE 1) THEN BEGIN ; If /NO_DISPLAY option was not set

    WINDOW, 4, XSIZE=1000, YSIZE=500

    device, get_decomposed=c_decomp ; Get current state of the decomposed flag
    device, decomposed=0
    loadct, 3 ; Load CT of 'Red Temperature'

    if(data_type eq 'FIS_SW') then begin

            ttitle = 'AKARI/FIS Image   '+file_header
            xyouts, 0.5, 0.95, ttitle, charsize=2.2, /normal, align=0.5

            leftmemo = 'grid:'+grid_str+'"' $
              +',  max:'+narrow_max_str+' [MJy/sr]' $
              +',  min:'+narrow_min_str+' [MJy/sr]'
            xyouts, 0.25, 0.03, leftmemo, charsize=1.2, /normal, align=0.5

            rightmemo = 'grid:'+grid_str+'"' $
              +',  max:'+wide_max_str+' [MJy/sr]' $
              +',  min:'+wide_min_str+' [MJy/sr]'
            xyouts, 0.75, 0.03, rightmemo, charsize=1.2, /normal, align=0.5

            contour, coad_det[1,*,*], im_ra, im_dec, nlevels=50, /fill, $
              xrange=[im_ra[N_ELEMENTS(im_ra)-1], im_ra[0]], $
              max_value=coad_narrow_max, min_value=coad_narrow_min, $
              title='N60', charsize=1.5, $
              xtitle='R.A. [deg]', ytitle='Dec [deg]', $
              position=[0.1,0.18,0.45,0.88], $
              /noerase

            contour, coad_det[0,*,*], im_ra, im_dec, nlevels=50, /fill, $
              xrange=[im_ra[N_ELEMENTS(im_ra)-1], im_ra[0]], $
              max_value=coad_wide_max, min_value=coad_wide_min, $
              title='Wide-S', charsize=1.5, $
              xtitle='R.A. [deg]', ytitle='Dec [deg]', $ 
              position=[0.6,0.18,0.95,0.88], $
              /noerase

        endif else begin

            ttitle = 'AKARI/FIS Image   '+file_header
            xyouts, 0.5, 0.95, ttitle, charsize=2.2, /normal, align=0.5

            leftmemo = 'grid:'+grid_str+'"' $
              +',  max:'+wide_max_str+' [MJy/sr]' $
              +',  min:'+wide_min_str+' [MJy/sr]'
            xyouts, 0.25, 0.03, leftmemo, charsize=1.2, /normal, align=0.5

            rightmemo = 'grid:'+grid_str+'"' $
              +',  max:'+narrow_max_str+' [MJy/sr]' $
              +',  min:'+narrow_min_str+' [MJy/sr]'
            xyouts, 0.75, 0.03, rightmemo, charsize=1.2, /normal, align=0.5

            contour, coad_det[0,*,*], im_ra, im_dec, nlevels=50, /fill, $
              xrange=[im_ra[N_ELEMENTS(im_ra)-1], im_ra[0]], $
              max_value=coad_wide_max, min_value=coad_wide_min, $
              title='Wide-L', charsize=1.5, $
              xtitle='R.A. [deg]', ytitle='Dec [deg]', $
              position=[0.1,0.18,0.45,0.88], $
              /noerase

            contour, coad_det[1,*,*], im_ra, im_dec, nlevels=50, /fill, $
              xrange=[im_ra[N_ELEMENTS(im_ra)-1], im_ra[0]], $
              max_value=coad_narrow_max, min_value=coad_narrow_min, $
              title='N160', charsize=1.5, $
              xtitle='R.A. [deg]', ytitle='Dec [deg]', $ 
              position=[0.6,0.18,0.95,0.88], $
              /noerase

        endelse


; Plot coadded number of samples

    !P.MULTI = [0, 2, 2]
    WSET, 0

    plot, coad_num0[0,*,*], yrange=[-1,21], PSYM=1, $
        title='initial coadded # (wide)', $
        xtitle='index', ytitle='COAD #'

    plot, coad_num0[1,*,*], yrange=[-1,21], PSYM=1, $
        title='initial coadded # (narrow)', $
        xtitle='index', ytitle='COAD #'

    sigmamemow = Strcompress(float(double(sigma)))+'-sigma regected sample # (wide)'
    sigmamemon = Strcompress(float(double(sigma)))+'-sigma regected sample # (narrow)'

    plot, coad_num0[0,*,*]-coad_num[0,*,*], yrange=[-1,11], PSYM=1, $
        title=sigmamemow, xtitle='index', ytitle='COAD #'

    plot, coad_num0[1,*,*]-coad_num[1,*,*], yrange=[-1,11], PSYM=1, $
        title=sigmamemon, xtitle='index', ytitle='COAD #'

    !P.MULTI = 0

    device, decomposed=c_decomp ; Restore decomposed flag

  ENDIF ; End of contour image and plot windows display



; Save IDL data
  if keyword_set(newmapping) then begin
      coad_file = file_header+'_img.sav'
  endif else begin
      if (scut ne 0) then begin ; in case of scut=ON
          if (sloop ne scan) then begin
              coad_file = file_header+'_img_'+snum+'.sav'
          endif
          if (sloop eq scan) then begin
              coad_file = file_header+'_img.sav'
          endif
      endif
      if (scut eq 0) then begin ; in case of scut=OFF
          coad_file = file_header+'_img.sav'
      endif
  endelse

  save, im_ra, im_dec, coad_det, coad_err, coad_num, data_type, $
    map_mean, dec_grid, ra_grid, filename=coad_file
  print, 'Output coadd image data in IDL format:', coad_file


; Save FITS data
  if keyword_set(newmapping) then begin
      coad_fits = file_header+'_img'
  endif else begin
      if (scut ne 0) then begin ; in case of scut=ON
          if (sloop ne scan) then begin
              coad_fits = file_header+'_img_'+snum
          endif
          if (sloop eq scan) then begin
              coad_fits = file_header+'_img'
          endif
      endif
      if (scut eq 0) then begin ; in case of scut=OFF
          coad_fits = file_header+'_img'
      endif
  endelse

  ss_coad_to_fits, im_ra, im_dec, coad_det, coad_err, coad_num, FILENAME=coad_fits
  print, 'Output coadd image data in FITS format:', coad_fits, '_[w|n].fits'


; Scan Cut Loop End
if keyword_set(newmapping) then goto, skipnewmap2
if (scut eq 0) then goto, skipscut3     ; in case of scut=OFF
endfor
skipscut3:
skipnewmap2:
skipmap2:

error_status = 0                ; set error status

; End of this pro
end
