;+
; NAME:
;   ss_flux_ave
;
; PURPOSE:
;   Calculate mean flux of line_fit data in given time range
;   Output calibration data files, i.e., CAL, Dark, Flat and CAL after effect tables
;
; CALLING SEQUENCE:
;
; MODIFICATION HISTORY:
;   Written by S. Matsuura, 2006/11/14
;
;   2006/12/12 changed notation of the output table from D11.7 to E12.5
;   2006/12/21 I.Yamamura plot window selection becomes more gentle
;   2006/12/29 Makiuti, S. Added NO_DISP option.
;   2007/01/24 S.Matsuura, changed sigma value 2 --> 3 for the bad data rejection.
;   2007/01/25 S.Matsuura, fixed bug for n=3.
;   2007/01/28 S.Matsuura, changed simple mean -> weighted mean.
;   2007/01/29 Makiuti, S. changed format code of E12.5 into E13.5 to avoid error on Windows
;   2007/01/30 S.Matsuura, handled blank data with zero value.
;
;-
pro ss_flux_ave, outfile, x_ave, coeff, error, n_pix, plot_pix, start_idx, end_idx, $
        NO_DISP=no_disp

; Check the parameters and initialize
  IF (n_elements(no_disp) EQ 0) THEN no_disp=0

; print input parameter
    print, 'Calculating mean flux ...'
    print, 'Start index:', start_idx, ' End index:', end_idx


; generate index for averaging line_fit data

    flux_mean = dblarr(3,n_pix)
    mean_idx = where( (x_ave ge start_idx) and (x_ave le end_idx), mean_idx_cnt)
    if (mean_idx_cnt gt 0) then begin

        x_val = x_ave[mean_idx]
        data = coeff[*,mean_idx,1]
        err = error[*,mean_idx,1]

        ; cut end point data
            n = n_elements(x_val)
            x_val = x_val[1:n-2]
            data = data[*,1:n-2]
            err = err[*,1:n-2]

    endif else begin

        n = 3
        x_val = dblarr(3)
        data = dblarr(n_pix,3)
        err = dblarr(n_pix,3)
        idx = where( min(x_ave-x_val), idx_cnt)
        x_val[0] = start_idx-1
        x_val[1] = start_idx
        x_val[2] = start_idx+1
        data[*,0] = coeff[*,idx,1]
        data[*,1] = coeff[*,idx,1]
        data[*,2] = coeff[*,idx,1]
        err[*,0] = error[*,idx,1]
        err[*,1] = error[*,idx,1]
        err[*,2] = error[*,idx,1]

    endelse

    n_data = n_elements(x_val)

; calculate mean flux / calculating median

    flux_mean0 = dblarr(3,n_pix)
    flux_stdev0 = dblarr(n_pix)

    for pix = 0, n_pix-1 do begin
        for j = 0, n_data-1 do begin
        if(data[pix,j] ne 0 and err[pix,j] gt 0) then begin
            flux_mean0[0,pix] = flux_mean0[0,pix] + data[pix,j]/err[pix,j]^2
            flux_mean0[1,pix] = flux_mean0[1,pix] + 1D/err[pix,j]^2
            flux_mean0[2,pix] = flux_mean0[2,pix] + 1
        endif
        endfor
        flux_stdev0[pix] = stddev(data[pix,*],/double,/NaN)
    endfor

    nz_pix0 = where(flux_mean0[1,*] gt 0, nz_pix_cnt0)
    if (nz_pix_cnt0 gt 0) then begin
	flux_mean0[0,nz_pix0] = flux_mean0[0,nz_pix0]/flux_mean0[1,nz_pix0]
	flux_mean0[1,nz_pix0] = sqrt(1D/flux_mean0[1,nz_pix0])
    endif


; re-calculate mean flux with 2-sigma rejection

    flux_mean = dblarr(3,n_pix)
    flux_stdev = dblarr(n_pix)

    for pix = 0, n_pix-1 do begin
        nrj_idx = where(abs(data[pix,*]-flux_mean0[0,pix]) lt 2*flux_stdev0[pix], nrj_cnt)
        if (nrj_cnt gt 0) then begin
        for j = 0, n_data-1 do begin
        if(data[pix,j] ne 0 and err[pix,j] gt 0) then begin
            flux_mean[0,pix] = flux_mean[0,pix] + data[pix,j]/err[pix,j]^2
            flux_mean[1,pix] = flux_mean[1,pix] + 1D/err[pix,j]^2
            flux_mean[2,pix] = flux_mean[2,pix] + 1
        endif
        endfor
        flux_stdev[pix] = stddev(data[pix,*],/double,/NaN)
        endif
    endfor

    nz_pix = where(flux_mean[1,*] gt 0, nz_pix_cnt)
    if (nz_pix_cnt gt 0) then begin
	flux_mean[0,nz_pix] = flux_mean[0,nz_pix]/flux_mean[1,nz_pix]
	flux_mean[1,nz_pix] = sqrt(1D/flux_mean[1,nz_pix])
    endif


  IF (no_disp NE 1) THEN BEGIN ; If /NO_DISP option was not set

; select window
;        window, 0
        WSET, 0
        pmulti_bk = !P.MULTI
        !P.MULTI = [0, 1, 2]

; plot temporal data

;   window, 0
    plot, x_val, data[plot_pix-1,*], title=outfile, $
        xtitle='# of samples', ytitle='output [V/s]', $
        ticklen=1.0, xgridstyle=1, ygridstyle=1
;       ticklen=1.0, xgridstyle=1, ygridstyle=1, $
;       position=[0.1,0.1,0.9,0.9]


; plot pixel data
    pix_idx = indgen(n_pix)
;   window, 1
    plot, pix_idx+1, flux_mean[0,*], $
        title=outfile, xtitle='# of pixel', ytitle='output [V/s]', $
        ticklen=1.0, xgridstyle=1, ygridstyle=1
;       ticklen=1.0, xgridstyle=1, ygridstyle=1, $
;       position=[0.1,0.1,0.9,0.9]

    errplot, pix_idx+1, $
        flux_mean[0,*] - flux_mean[1,*], $
        flux_mean[0,*] + flux_mean[1,*]

; reset window setting
        !P.MULTI = pmulti_bk

  ENDIF ; End of Plots dispplay


; write flux table

    openw, unit, outfile, /get_lun

    for pix = 0, n_pix-1 do begin
        printf, unit, pix+1, flux_mean[0,pix], flux_mean[1,pix], flux_mean[2,pix], $
          format = '(I3,2X,E13.5,2X,E13.5,2X,E13.5)'
    endfor

    free_lun, unit


end
