;+
; NAME:
;   SS_MAKE_MAP
;
; PURPOSE:
;   Main program of slow-scan data analysing tool
;   Call responsivity calibration/correction procedures
;     - responsivity correction and flat fielding
;     - background subtraction by filtering and estimation of stray light
;     - making coadd image
;   Output time series, surface brightness data and map data in MJy/sr 
;
; INPUT: '*_pr.sav' (IDL format data)
;
; OUTPUT:
;    Time series processed data, RADEC position (Text): '*_fit_ar.txt'
;    Time series processed data, RADEC position, and parameters (IDL): '*_ar.sav'
;    * "ar" denotes After-Reduction
; - Indirect OUTPUT via "ss_coad_image.pro":
;    Co-add image (PS): '*_img_[w|n].ps', same as shown on the screen
;    Co-add image data, RADEC, and parameters (IDL): '*_img.sav'
;    Co-add image data, RADEC (FITS): '*_img_[w|n].fits'
;
; CALLING SEQUENCE:
;
; MODIFICATION HISTORY:
;   2006/12/27 Written by S. Matsuura, based on "ss_line_fit.pro"
;   2006/12/29 Makiuti, S., Added /NO_DISPLAY options.
;   2006/12/29 S. Matsuura, included more variables in IDL-format output "*_ar.sav".
;   2006/12/29 S. Matsuura, cal_aft_cor option is automatically chosen when 'AOT' is 'PV_*',
;   2007/01/05 Makiuti, S., Improved error handling. Added ERROR_STATUS option argument to 
;              return error status.
;   2007/01/09 S.Matsuura, sigma for bad data rejection can be specified.
;              Time range for mapping data can be specified as between t_start and t_end [s].
;   2007/01/17 M.Shirahata, added /sky_sub option, and deleted /mapping option.
;   2007/01/18 S.Matsuura, fixed bug of ,/median option.
;   2007/01/18 M.Shirahata, set the default value of t_start(=630@P12_*, 750@PV_*) and t_end(=1350).
;   2007/01/23 M.Shirahata, change output file names (.dat --> .txt)
;   2007/01/23 S.Matsuura, included bad data flag found in the flat fielding process.
;   2007/01/30 S.Matsuura, handled blank data with zero value.
;   2007/01/31 M.Shirahata, fixed bug of saving '_ar.sav' data.
;                           change how to treat the t_start and t_end options.
;   2007/02/01 M.Shirahata, change option name (median_flt --> median_filter)
;   2007/02/03 S.Matsuura, added new parameter of bad data rejection threshold.
;   2007/02/03 S.Matsuura, cleaned messy code of zero data handling.
;   2007/02/05 S.Matsuura, added "width_median" option as window size of median filter.
;   2007/02/06 M.Shirahata, add scut option.
;   2007/02/08 S.Matsuura, added grid size input option.
;-
pro ss_make_map, infile, $
        MASK_OFF = mask_off, $
        RESP_COR_OFF = resp_cor_off, $
        SL_RMV = sl_rmv, $
        MEDIAN_FILTER = median_filter, $
        LOCAL_FLAT = local_flat, $
        SKY_SUB = sky_sub, $
        im_ra, im_dec, coad_det, coad_err, coad_num, $
        NO_DISPLAY = no_display, $
        ERROR_STATUS = error_status, $
	SIGMA = sigma, $
	T_START = t_start, T_END = t_end, $
	BAD_THRSHLD = bad_thrshld, $
	WIDTH_MEDIAN = width_median, $
        SCUT = scut, GRID_SW=grid_sw, GRID_LW=grid_lw

; make filename
    ; remove '_pr.sav'
    file_head_len = strlen(infile) - 7
    file_header = strmid(infile, 0, file_head_len)
    print, file_header

; Restore IDL data
    restore, filename=infile
    print, 'Read pre-reduction linear-fit data in IDL format:', infile
    help, x_ave, q_mean, coeff, error, eg_gl_cnt, drv_mask_tot, cal_flg_tot, $
        data_type, AOT, RST, n_int, rcount, rst_int, $
        ramp_cor_off, glitch_rmv_off, aocu_ads

; data type
    if(data_type eq 'FIS_SW') then begin
        n_pix = 100 ; total pix number
        w_pix = 60  ; wide band
        smpl_freq = 25.28D ; [Hz]
        pix_omega = 2.1154e-8 ; str
        capacitance = 7.0D ; pF
        spk_thrld = 4.0D
        bad_flg = intarr(n_pix)
        bad_flg[0] = 1  ; ch1
        bad_flg[10] = 1 ; ch11
        bad_flg[70] = 1 ; ch71
        bad_flg[84] = 1 ; ch85
        bad_flg[97] = 1 ; ch98
    endif else begin
        n_pix = 75  ;total pix number
        w_pix = 45  ; wide band
        smpl_freq = 16.86D ; [Hz]
        pix_omega = 5.8761e-8 ; str
        capacitance = 3.0D ; pF
        spk_thrld = 5.0D
        bad_flg = intarr(n_pix)
        bad_flg[1] = 1  ; ch2
        bad_flg[2] = 1  ; ch3
        bad_flg[17] = 1 ; ch18
        bad_flg[32] = 1 ; ch33
        bad_flg[41] = 1 ; ch42
        bad_flg[47] = 1 ; ch48
        bad_flg[62] = 1 ; ch63
        bad_flg[69] = 1 ; ch70
   endelse


; Correction of responsivity drift after pulsed cal-lamp
if (AOT eq 'PV_FIS01' or AOT eq 'PV_FIS02') then begin

    ; detect bad data with zero value
    fisv_bad_idx = where(coeff eq 0 and error eq 0, fisv_bad_cnt)

    ; read cal after effect correction data
    cal_cor_file = file_header + '_aftcal.txt'
    openr,lun, cal_cor_file, /Get_LUN, ERROR=oper
    if(oper eq 0) then begin
        print, 'Read cal after effect correction parameter: ', cal_cor_file
    endif else begin
        print, 'Cannot find cal after effect correction parameter file : ', cal_cor_file
        error_status = 1 ; set error status
        return
    endelse

    cal_cor = dblarr(30*n_int, n_pix)
    READF, lun, cal_cor
    Free_lun,lun

    ; additive method to correct
    cal_flg_idx = where(cal_flg_tot ne 0, cnt)
    print, 'cal flag count: ', cnt

    if(cnt gt 0) then begin
    for pix = 0, n_pix-1 do begin
        for j = 0, 10*n_int-1 do begin
            coeff[pix,cal_flg_idx+j,1] = coeff[pix,cal_flg_idx+j,1] - cal_cor[j,pix]
            ;error[pix,cal_flg_idx+j,1] = sqrt(error[pix,cal_flg_idx+j,1]^2 + cal_cor[10+j,pix]^2)
        endfor
    endfor
    endif

    ; bad data is reset to zero
    if (fisv_bad_cnt gt 0) then begin
	coeff[fisv_bad_idx] = 0
	error[fisv_bad_idx] = 0
    endif

endif


; mask shutter and cal-lamp data
if keyword_set(mask_off) then begin

    x = x_ave
    n_line = n_elements(x)

    det = dblarr(n_pix,n_line)
    err = dblarr(n_pix,n_line)

    det[*,*] = coeff[*,*,1]
    err[*,*] = error[*,*,1]

endif else begin

    drv_mask_idx = where(drv_mask_tot eq 0, drv_mask_cnt)
    x = x_ave[drv_mask_idx]
    n_line = n_elements(x)

    det = dblarr(n_pix,n_line)
    err = dblarr(n_pix,n_line)

    for j = 0, n_line-1 do begin
        det[*,j] = coeff[*,drv_mask_idx[j],1]
        err[*,j] = error[*,drv_mask_idx[j],1]
        q_mean[*,j] = q_mean[*,drv_mask_idx[j]]
        eg_gl_cnt[*,j] = eg_gl_cnt[*,drv_mask_idx[j]]
    endfor

    det = det[*,0:n_line-1]
    err = err[*,0:n_line-1]
    q_mean = q_mean[*,0:n_line-1]
    eg_gl_cnt = eg_gl_cnt[*,0:n_line-1]

endelse


; Dark subtraction and Responsivity correction
if keyword_set(resp_cor_off) then goto, resp_cor_skip

    ; detect bad data with zero value
    fisv_bad_idx = where(det eq 0 and err eq 0, fisv_bad_cnt)

    ; set bad pixel rejection threshold
    if (n_elements(bad_thrshld) eq 0) then bad_thrshld = 10.0D

    ; Responsivity calculation procedure
    if(keyword_set(local_flat) eq 1) then begin
        ss_resp_corr, file_header, data_type, AOT, RST, dark, sky_bg, $
		responsivity, bad_flg, bad_thrshld, /local_flat
    endif else begin
        ss_resp_corr, file_header, data_type, AOT, RST, dark, sky_bg, $
		responsivity, bad_flg, bad_thrshld
    endelse

    fine_pix = where(bad_flg eq 0 and responsivity[0,*] gt 0, fine_pix_cnt) 

    if(keyword_set(sky_sub) eq 1) then begin
        ; Responsivity correction & sky subtraction
        if(data_type eq 'FIS_SW') then begin
            for j = 0, n_line-1 do begin
                det[fine_pix,j] = (det[fine_pix,j]-dark[1,fine_pix]-sky_bg[0,fine_pix])/responsivity[0,fine_pix] ; [MJy/sr]
                err[fine_pix,j] = sqrt(err[fine_pix,j]^2+sky_bg[1,fine_pix]^2)/responsivity[0,fine_pix] ; [MJy/sr]
            endfor
        endif else begin ; in case of 'FIS_LW'
            for j = 0, n_line-1 do begin
                det[fine_pix,j] = (det[fine_pix,j]-sky_bg[0,fine_pix])/responsivity[0,fine_pix] ; [MJy/sr]
                err[fine_pix,j] = sqrt(err[fine_pix,j]^2+sky_bg[1,fine_pix]^2)/responsivity[0,fine_pix] ; [MJy/sr]
            endfor
        endelse
    endif else begin
        ; Responsivity correction
        for j = 0, n_line-1 do begin
            det[fine_pix,j] = (det[fine_pix,j]-dark[1,fine_pix])/responsivity[0,fine_pix] ; [MJy/sr]
            err[fine_pix,j] = err[fine_pix,j]/responsivity[0,fine_pix] ; [MJy/sr]
       	endfor
    endelse

    ; bad data is reset to zero
    if (fisv_bad_cnt gt 0) then begin
	det[fisv_bad_idx] = 0
	err[fisv_bad_idx] = 0
    endif

resp_cor_skip:


; calculate pixel position in RADEC and draw image of each array column
root = GETENV('FISDR_ROOT')
if(data_type eq 'FIS_SW') then begin
    pix_pos_file = root+'/slowscan/lib/FIS-SW_pix_pos_distort1.dat'
endif else begin
    pix_pos_file = root+'/slowscan/lib/FIS-LW_pix_pos_distort1.dat'
endelse

ss_q_to_radec, q_mean, data_type, pix_pos_file, ra, dec


; remove DC components

; detect bad data with zero value
    fisv_bad_idx = where(det eq 0 and err eq 0, fisv_bad_cnt)

; model and subtract the stray light, the earth shine reflected at telescope baffle
; *** under construction ***
if keyword_set(sl_rmv) then begin

    ; using stray light model
    ss_stray_light, x, ra, dec, det, err, eg_gl_cnt, file_header, data_type, $
      strm_det, strm_err
    det = strm_det
    err = strm_err

    ; using simple rejection of offset component
    det_ave = dblarr(2,n_line)
    fine_pix = where(bad_flg eq 0, fine_pix_cnt) 
    fine_w_idx = where(fine_pix lt w_pix, fine_w_cnt)
    fine_n_idx = where(fine_pix ge w_pix, fine_n_cnt)
    for j = 0, n_line-1 do begin
        det_ave[0,j] = median(det[fine_pix[fine_w_idx],j])
        det_ave[1,j] = median(det[fine_pix[fine_n_idx],j])
        det[fine_pix[fine_w_idx],j] = det[fine_pix[fine_w_idx],j] - det_ave[0,j]
        det[fine_pix[fine_n_idx],j] = det[fine_pix[fine_n_idx],j] - det_ave[1,j]
    endfor

endif

; median smooth filter
if keyword_set(median_filter) then begin

    if (n_elements(width_median) eq 0) then width_median = 10*n_int
    det_med = dblarr(n_pix,n_line+2*width_median)
    det_p = dblarr(n_line+2*width_median)
    err_p = dblarr(n_line+2*width_median)
    for pix = 0, n_pix-1 do begin
	if (bad_flg[pix] eq 0) then begin
	    det_p[width_median:n_line+width_median-1] = det[pix,0:n_line-1]
	    det_p[0:width_median-1] = det[pix,0]
	    det_p[n_line+width_median:n_line+2*width_median-1] = det[pix,n_line-1]
	    err_p[width_median:n_line+width_median-1] = err[pix,0:n_line-1]
	    err_p[0:width_median-1] = err[pix,0]
	    err_p[n_line+width_median:n_line+2*width_median-1] = err[pix,n_line-1]
	    nz_idx = where(det_p ne 0 and err_p ne 0, nz_idx_cnt)
	    if (nz_idx_cnt gt 0) then begin
	    det_med[pix,nz_idx] = median(det_p[nz_idx],width_median,/double)
	    ; other types of filter
	    ; det_med[pix,nz_idx] = smooth(det_p[nz_idx],width_median,/edge_truncate)
	    ; filter = digital_filter(0.0,1.0/width_median,width_median,50,/double)
	    ; det_med[pix,nz_idx] = convol(det_p[nz_idx],filter)
	    endif
	endif
	det_p[*] = 0D
	err_p[*] = 0D
    endfor
    det_med = det_med[*,width_median:n_line+width_median-1]
    det = det - det_med

    det = det[*,0:n_line-1]
    err = err[*,0:n_line-1]
    q_mean = q_mean[*,0:n_line-1]
    eg_gl_cnt = eg_gl_cnt[*,0:n_line-1]

endif


; zero value reset for bad data

    ; fisv bad data -> 0
    if (fisv_bad_cnt gt 0) then begin
	det[fisv_bad_idx] = 0
	err[fisv_bad_idx] = 0
    endif

    ; bad pixel data -> 0
    bad_pix = where(bad_flg gt 0, bad_pix_cnt)
    if (bad_pix_cnt gt 0) then begin
	det[bad_pix,*] = 0
	err[bad_pix,*] = 0
    endif

    ; glitch affected data -> 0
    ; gl_bad_idx = where(eg_gl_cnt gt 0, gl_bad_cnt)
    ; if (gl_bad_cnt gt 0) then begin
    ;    det[gl_bad_idx] = 0
    ;    err[gl_bad_idx] = 0
    ; endif


; make filename
dat_file = '_fit_ar.txt'
filename = file_header + dat_file
print, 'Output text data:', filename


; write time_series output TXT file
openw, unit0, filename, /get_lun

if(data_type eq 'FIS_SW') then begin
    for j=0L, n_line-1L do begin
        printf,unit0,x[j],ra[*,j],dec[*,j],det[*,j],err[*,j],eg_gl_cnt[*,j], $
          format='(F16.3,2X,100(F12.7,2X),100(F12.7,2X),100(F12.7,2X),100(F12.7,2X),100(F12.7,2X))'
    endfor
endif else begin
    for j=0L, n_line-1L do begin
        printf,unit0,x[j],ra[*,j],dec[*,j],det[*,j],err[*,j],eg_gl_cnt[*,j], $
          format='(F16.3,2X,75(F12.7,2X),75(F12.7,2X),75(F12.7,2X),75(F12.7,2X),75(F12.7,2X))'
    endfor
endelse

free_lun, unit0


; Save IDL data
save, x, ra, dec, det, err, det_med, eg_gl_cnt, $
  data_type, AOT, RST, rst_int, n_int, bad_flg, $
  ramp_cor_off, glitch_rmv_off, aocu_ads, $
  filename=file_header+'_ar.sav'
print, 'Output IDL-format data:', file_header+'_ar.sav'


; make coadd image
IF ~keyword_set(scut) THEN scut = 0

ss_coad_image, x, ra, dec, det, err, eg_gl_cnt, file_header, $
  data_type, AOT, RST, rst_int, n_int, bad_flg, $
  im_ra, im_dec, coad_det, coad_num, coad_err, $
  NO_DISPLAY = no_display, ERROR_STATUS = error_status, $
  SIGMA = sigma, T_START = t_start, T_END = t_end, SCUT=scut, $
  GRID_SW = grid_sw, GRID_LW = grid_lw


if (error_status EQ 1) then begin
  message, 'SS_COAD_IMAGE returned error. Process stopped!', /continue
  return
endif

error_status = 0                ; set error status
MESSAGE, 'FINISH.',/CONTINUE

end
