;+
; NAME:
;   ss_stray_light
;
; PURPOSE:
;   model and remove the stray light of earth shine
;
; CALLING SEQUENCE:
;
; MODIFICATION HISTORY:
;   Written by S. Matsuura, 2006/09/05
;   2006/10/10 no use of data points affected by glitches
;   2006/10/19 usable for any sky by including ra dec to input parameter
;              save fitting parameter
;   2006/10/25 I.Yamamura renamed ss_stray_light
;   2006/12/21 I.yamamura replace WINDOW with WSET
;   2007/04/21 I.yamamura move ss_strayfunc before ss_stray_light so that
;              the subroutine is compiled at the same time.
;              array index adjusted to avoid crash around line 149
;-

;
; subroutine called from curvefit()
;
pro ss_strayfunc, x_pix, a, func, pder

	PI      = 3.1415926535897932385D
	RAD2DEG = 57.295779513082320877D   ; [deg/rad]

	re = 6400.0 ; erth radius in km
	hs = 700.0 ; satellite altitude in km
	lambda = a[3] ; e-lon in rad
	alpha = (23.4*sin(lambda-PI/2.0)+8.0)/RAD2DEG ; 23.4 + inclination - 90.0
	beta = a[4] ; e-lat in rad
	omega = PI/2.0 - beta
	fai = acos(re/(re+hs)) ; earth limb angle

	; theta: variable, satellite phase angle in orbit from pointing direction
	time_offset = 0.0 ; in second
	time_mid = 885.0 ; in second
	theta = omega - (1.047197551e-3)*(x_pix-(time_mid-time_offset))
	gamma = acos(cos(theta)*cos(alpha)*cos(omega)+sin(theta)*sin(omega))
	delta = fai + PI/2.0 - gamma ; earth avoidance angle

	bx = exp(-1.0*a[1]*delta^2)
	func = a[0] * bx + a[2]

	pder = [ [bx], [a[0]*(-1.0*delta^2)*bx], $
		[replicate(1.0, n_elements(x_pix))], $
		[replicate(0.0, n_elements(x_pix))], $
		[replicate(0.0, n_elements(x_pix))] ]

end
;
; ss_Stray_Light main routine starts here.
;
pro ss_stray_light, x, ra, dec, det, err, eg_gl_cnt, file_header, data_type, $
		strm_det, strm_err


PI      = 3.1415926535897932385D
RAD2DEG = 57.295779513082320877D   ; [deg/rad]

; data type

	if(data_type eq 'FIS_SW') then begin
		n_pix = 100 ; total pix number
		w_pix = 60  ; wide band
		smpl_freq = 25.28D; [Hz]
		bad_flg = intarr(n_pix)
		bad_flg[0] = 1  ; ch1
		bad_flg[10] = 1 ; ch11
		bad_flg[70] = 1 ; ch71
		bad_flg[84] = 1 ; ch85
		bad_flg[97] = 1 ; ch98
	endif else begin
		n_pix = 75  ;total pix number
		w_pix = 45  ; wide band
		smpl_freq = 16.86D; [Hz]
		bad_flg = intarr(n_pix)
		bad_flg[1] = 1  ; ch2
		bad_flg[2] = 1 ; ch3
		bad_flg[17] = 1 ; ch18
		bad_flg[32] = 1 ; ch33
		bad_flg[41] = 1 ; ch42
		bad_flg[47] = 1 ; ch48
		bad_flg[62] = 1 ; ch63
		bad_flg[69] = 1 ; ch70
	endelse


; Mid time data

	mid_idx = fix(n_elements(det)/n_pix/2.0)
	ave_width = fix(n_elements(det)/n_pix/10.0)
	ave_st_idx = mid_idx - ave_width
	ave_en_idx = mid_idx + ave_width


; Ecliptic coordinate of target, to be calculated from RA and DEC

	euler, ra, dec, lambda, beta, 3
	lambda_ave = mean(lambda[ave_st_idx:ave_en_idx])
	beta_ave = mean(beta[ave_st_idx:ave_en_idx])
	print, 'Mean ecliptic longitude [deg]:', lambda_ave
	print, 'Mean ecliptic latitude [deg]:', beta_ave

	lambda_ave = lambda_ave/RAD2DEG
	beta_ave = beta_ave/RAD2DEG


; Median smoothing before fitting

	width = 10
	;det_median = median(det, width, /double, dimension = 2)
	det_median = det


; gaussian beam fitting

	x_time = x / smpl_freq
	strm_det = det
	strm_err = err
	str_light = det
	bkgnd = dblarr(n_pix)
	sl_model = dblarr(n_pix,8)

;	window, 0
	WSET, 0

	for pix = 0, n_pix-1 do begin
	if(bad_flg[pix] ne 1) then begin

		gl_mask_idx = where(eg_gl_cnt[pix,*] eq 0, gl_mask_cnt)
		if(gl_mask_cnt gt 0) then begin
			x_pix = x_time[gl_mask_idx]
			det_pix = det_median[pix, gl_mask_idx]
			err_pix = err[pix, gl_mask_idx]
		endif else begin
			x_pix = x_time
			det_pix = det_median[pix, *]
			err_pix = err[pix, *]
		endelse

		print, 'ch:', pix+1
		ploterr, x_pix, det_pix, err_pix

		; determine initial value for fitting
                  ; 07/04/21 I.Yamamura to avoid crash
                ave_st_idx_l = ave_st_idx
                ave_en_idx_l = ave_en_idx
                if (ave_st_idx_l LT 0) then ave_st_idx = 0
                if (ave_en_idx_l GE N_ELEMENTS(det_pix)) then ave_en_idx_l = N_ELEMENTS(det_pix) - 1

		det_ave = median(det_pix[ave_st_idx_l:ave_en_idx_l])
		print, 'sky brightness at the minimum of stray light: ', det_ave
		det_stdev = stddev(det_pix[ave_st_idx_l:ave_en_idx_l])
		print, 'std_dev at the minimum: ', det_stdev

		; remove dat at boundary 
		n_fit = n_elements(x_pix)
		fit_st = 10*width
		fit_en = n_fit - 10*width
		x_pix = x_pix[fit_st:fit_en]
		det_pix = det_pix[fit_st:fit_en]
		err_pix = err_pix[fit_st:fit_en]

		weights = replicate(1.0D/det_stdev^2, n_elements(x_pix))
		a = [det_ave*100.0, 2.5, det_ave, lambda_ave, beta_ave]
		fita = [1, 1, 1, 0, 0]

		coefs = curvefit(x_pix, det_pix, weights, a, sigma, $
			/double, FITA=fita, FUNCTION_NAME='ss_strayfunc', $
			CHISQ=chisq, STATUS=status, ITMAX=50)

		oplot, x_pix, coefs

		print, 'function parameters:', a
		print, 'parameter errors   :', sigma
		print, 'chi-square:', chisq
		print, 'Fitting status (0:successful, 1|2:failed) :', status

		n = n_elements(x_time)
		for i = 0, n-1 do begin
			ss_strayfunc, x_time[i], a, func, pder
			str_light[pix,i] = func - a[2]
		endfor

		for j = 0, 2 do begin
			sl_model[pix,j] = a[j]
			sl_model[pix,j+3] = sigma[j]
		endfor
		sl_model[pix,6] = chisq
		sl_model[pix,7] = status

	endif
	endfor

	strm_det = det - str_light


; save result of fitting

	dat_file = '_slpara.txt'
	filename = file_header + dat_file
	print, 'Save stray-light fitting result:', filename

	openw, unit0, filename, /get_lun
	for pix = 0, n_pix-1 do begin
		printf, unit0, sl_model[pix,*], format='(8(F12.3,2X))'
	endfor
	free_lun, unit0

end
